from typing import Optional, Literal, List

from .basemodel import Base
from .user import User


class Covers(Base):
    cover: str
    card: str
    list: str
    slimcover: str


class BeatmapsetCompact(Base):
    artist: str
    artist_unicode: str
    covers: Covers
    creator: str
    favourite_count: int
    id: int
    nsfw: bool
    play_count: int
    preview_url: str
    source: str
    title: str
    title_unicode: str
    beatmapset_id: Optional[int]
    user_id: int
    status: str
    video: bool
    converts: Optional[str]
    description: Optional[str]
    has_favourited: Optional[bool]
    language: Optional[str]
    user: Optional[User]
    total_length: Optional[int]


class Beatmapset(BeatmapsetCompact):
    bpm: Optional[float]
    can_be_hyped: Optional[bool]
    ranked: Optional[int]
    ranked_date: Optional[str]
    tags: Optional[str]


class BeatmapCompact(Base):
    beatmapset_id: int
    difficulty_rating: float
    id: int
    mode: Literal['fruits', 'mania', 'osu', 'taiko']
    status: str
    total_length: int
    user_id: int
    version: str
    beatmapset: Optional[Beatmapset]
    checksum: Optional[str]
    max_combo: Optional[int]
    version: str


class Beatmap(BeatmapCompact):
    accuracy: float
    ar: float
    bpm: Optional[float]
    convert: bool
    count_circles: int
    count_sliders: int
    count_spinners: int
    cs: float
    deleted_at: Optional[str]
    drain: float
    hit_length: int
    is_scoreable: bool
    last_updated: str
    mode_int: int
    passcount: int
    playcount: int
    ranked: int
    url: str


class BeatmapDifficultyAttributes(Base):
    max_combo: int
    star_rating: float


class BackgroundsAttributes(Base):
    url: str
    user: dict


class SeasonalBackgrounds(Base):
    ends_at: str
    backgrounds: List[BackgroundsAttributes]



from typing import Optional
from tortoise import fields, Model


class UserData(Model):
    id: int = fields.IntField(pk=True, generated=True, auto_increment=True)
    """自增主键"""
    user_id: int = fields.BigIntField()
    """用户id"""
    osu_id: int = fields.IntField()
    """osu id"""
    osu_name: str = fields.TextField()
    """osu 用户名"""
    osu_mode: int = fields.IntField()
    """osu 模式"""

    class Meta:
        table = 'User'
        indexes = ('user_id',)


class InfoData(Model):
    id: int = fields.IntField(pk=True, generated=True, auto_increment=True)
    """自增主键"""
    osu_id: int = fields.IntField()
    """osu id"""
    c_rank: Optional[int] = fields.IntField(null=True)
    """国家排名"""
    g_rank: Optional[int] = fields.IntField(null=True)
    """世界排名"""
    pp: float = fields.FloatField()
    """pp"""
    acc: float = fields.FloatField()
    """acc"""
    pc: int = fields.IntField()
    """游戏次数"""
    count: int = fields.IntField()
    """打击note数"""
    osu_mode: int = fields.IntField()
    """osu 模式"""
    date = fields.DateField()

    class Meta:
        table = 'Info'
        indexes = ('id',)



from io import BytesIO, TextIOWrapper
from typing import Union, Optional

import re
from pathlib import Path

from nonebot import get_driver
from nonebot.log import logger

from .config import Config
from .api import sayo_api, safe_async_get
from .schema import SayoBeatmap, User, Badge

plugin_config = Config.parse_obj(get_driver().config.dict())


osufile = Path(__file__).parent / 'osufile'
map_path = Path() / 'data' / 'osu' / 'map'
user_cache_path = Path() / 'data' / 'osu' / 'user'
badge_cache_path = Path() / 'data' / 'osu' / 'badge'
chimu_api = 'https://api.chimu.moe/v1/download/'
kitsu_api = 'https://kitsu.moe/api/d/'
sayobot_api = 'https://txy1.sayobot.cn/beatmaps/download/novideo/'
api_ls = [sayobot_api, kitsu_api, chimu_api]
download_api = api_ls[plugin_config.osz_mirror]

if not map_path.exists():
    map_path.mkdir(parents=True, exist_ok=True)
if not user_cache_path.exists():
    user_cache_path.mkdir(parents=True, exist_ok=True)
if not badge_cache_path.exists():
    badge_cache_path.mkdir(parents=True, exist_ok=True)


async def download_map(setid: int) -> Optional[Path]:
    url = download_api + str(setid)
    logger.info(f'开始下载地图: <{setid}>')
    req = await safe_async_get(url)
    filename = f'{setid}.osz'
    filepath = map_path.parent / filename
    with open(filepath, 'wb') as f:
        f.write(req.read())
    logger.info(f'地图: <{setid}> 下载完毕')
    return filepath


async def download_tmp_osu(map_id):
    url = f'https://osu.ppy.sh/osu/{map_id}'
    logger.info(f'开始下载谱面: <{map_id}>')
    req = await safe_async_get(url)
    filename = f'{map_id}.osu'
    filepath = map_path / filename
    chunk = req.read()
    with open(filepath, 'wb') as f:
        f.write(chunk)
    return filepath


async def download_osu(set_id, map_id):
    url = f'https://osu.ppy.sh/osu/{map_id}'
    logger.info(f'开始下载谱面: <{map_id}>')
    req = await safe_async_get(url)
    filename = f'{map_id}.osu'
    filepath = map_path / str(set_id) / filename
    chunk = req.read()
    with open(filepath, 'wb') as f:
        f.write(chunk)
    return filepath


async def get_projectimg(url: str):
    if 'avatar-guest.png' in url:
        url = 'https://osu.ppy.sh/images/layout/avatar-guest.png'
    req = await safe_async_get(url)
    if req.status_code == 403:
        return osufile / 'work' / 'mapbg.png'
    data = req.read()
    im = BytesIO(data)
    return im


def re_map(file: Union[bytes, Path]) -> str:
    if isinstance(file, bytes):
        text = TextIOWrapper(BytesIO(file), 'utf-8').read()
    else:
        with open(file, 'r', encoding='utf-8') as f:
            text = f.read()
    res = re.search(r'\d,\d,\"(.+)\"', text)
    bg = 'mapbg.png' if not res else res.group(1).strip()
    return bg


async def get_map_id(file: Path) -> str:
    with open(file, 'r', encoding='utf-8') as f:
        text = f.read()
    if res := re.search(r'BeatmapID:(\d*)', text):
        return res.group(1).strip()
    res = re.search(r'Version:(.*)\n', text)
    version = res.group(1).strip()
    data = await sayo_api(int(file.parent.name))
    if isinstance(data, str):
        raise Exception(data)
    sayo_info = SayoBeatmap(**data)
    if sayo_info.status == -1:
        raise Exception('未能在sayobot查找到地图信息')
    for map_data in sayo_info.data.bid_data:
        if map_data.version == version:
            return str(map_data.bid)


# 当不存在缓存文件时，建立缓存文件
async def make_user_cache_file(info: User):
    path = user_cache_path / str(info.id)
    path.mkdir()
    user_header = await get_projectimg(info.cover_url)
    user_icon = await get_projectimg(info.avatar_url)
    with open(path / 'header.png', 'wb') as f:
        f.write(user_header.getvalue())
    with open(path / 'icon.png', 'wb') as f:
        f.write(user_icon.getvalue())


async def make_badge_cache_file(badge: Badge):
    path = badge_cache_path / f'{hash(badge.description)}.png'
    badge_icon = await get_projectimg(badge.image_url)
    with open(path, 'wb') as f:
        f.write(badge_icon.getvalue())


# 保存个人信息界面背景
async def save_info_pic(user: str, url):
    path = user_cache_path / user
    if not path.exists():
        path.mkdir()
    req = await safe_async_get(url)
    with open(path / 'info.png', 'wb') as f:
        f.write(BytesIO(req.content).getvalue())



from io import BytesIO
from typing import Union
from PIL import Image, UnidentifiedImageError

from ..api import osu_api, get_map_bg
from ..file import download_osu, re_map, map_path


async def get_bg(mapid: Union[str, int]) -> Union[str, BytesIO]:
    info = await osu_api('map', map_id=mapid)
    if not info:
        return '未查询到该地图'
    elif isinstance(info, str):
        return info
    setid: int = info['beatmapset_id']
    set_path = map_path / str(setid)
    if not set_path.exists():
        set_path.mkdir(parents=True, exist_ok=True)
    osu = map_path / str(setid) / f"{mapid}.osu"
    if not osu.exists():
        await download_osu(setid, mapid)
    cover = re_map(osu)
    cover_path = map_path / str(setid) / cover
    if not cover_path.exists():
        bg = await get_map_bg(setid, cover)
        with open(cover_path, 'wb') as f:
            f.write(bg.getvalue())
    try:
        img = Image.open(cover_path)
    except UnidentifiedImageError:
        cover_path.unlink()
        return '暂时无法下载背景图片＞︿＜'
    byt = BytesIO()
    img.save(byt, 'png')
    return byt



import math
import os
import random

import numpy as np
from io import BytesIO
from typing import Optional, Union
from PIL import ImageDraw, UnidentifiedImageError, ImageEnhance, ImageFilter
from matplotlib.figure import Figure

from ..file import get_projectimg, user_cache_path
from ..schema import SeasonalBackgrounds, User
from ..api import get_seasonal_bg, safe_async_get

from .static import *


def image2bytesio(pic: Image):
    byt = BytesIO()
    pic.save(byt, "png")
    return byt


def draw_fillet(img, radii):
    # 画圆（用于分离4个角）
    circle = Image.new('L', (radii * 2, radii * 2), 0)  # 创建一个黑色背景的画布
    draw = ImageDraw.Draw(circle)
    draw.ellipse((0, 0, radii * 2, radii * 2), fill=255)  # 画白色圆形
    # 原图
    img = img.convert("RGBA")
    w, h = img.size
    # 画4个角（将整圆分离为4个部分）
    alpha = Image.new('L', img.size, 255)
    alpha.paste(circle.crop((0, 0, radii, radii)), (0, 0))  # 左上角
    alpha.paste(circle.crop((radii, 0, radii * 2, radii)), (w - radii, 0))  # 右上角
    alpha.paste(circle.crop((radii, radii, radii * 2, radii * 2)), (w - radii, h - radii))  # 右下角
    alpha.paste(circle.crop((0, radii, radii, radii * 2)), (0, h - radii))  # 左下角
    # 白色区域透明可见，黑色区域不可见
    img.putalpha(alpha)
    return img

def draw_fillet2(img, radii):
    # 画圆（用于分离4个角）
    circle = Image.new('L', (radii * 2, radii * 2), 0)  # 创建一个黑色背景的画布
    draw = ImageDraw.Draw(circle)
    draw.ellipse((0, 0, radii * 2, radii * 2), fill=255)  # 画白色圆形
    # 原图
    img = img.convert("RGBA")
    w, h = img.size
    # 创建一个透明度为50%的图像
    enhancer = ImageEnhance.Brightness(img)
    img = enhancer.enhance(0.5)
    # 画4个角（将整圆分离为4个部分）
    alpha = Image.new('L', img.size, 255)
    alpha.paste(circle.crop((0, 0, radii, radii)), (0, 0))  # 左上角
    alpha.paste(circle.crop((radii, 0, radii * 2, radii)), (w - radii, 0))  # 右上角
    alpha.paste(circle.crop((radii, radii, radii * 2, radii * 2)), (w - radii, h - radii))  # 右下角
    alpha.paste(circle.crop((0, radii, radii, radii * 2)), (0, h - radii))  # 左下角
    # 高斯模糊效果
    img = img.filter(ImageFilter.GaussianBlur(radius=2))
    # 白色区域透明可见，黑色区域不可见
    img.putalpha(alpha)  
    return img

def info_calc(n1: Optional[float], n2: Optional[float], rank: bool = False, pp: bool = False):
    if not n1 or not n2:
        return '', 0
    num = n1 - n2
    if num < 0:
        if rank:
            op, value = '↑', num * -1
        elif pp:
            op, value = '↓', num * -1
        else:
            op, value = '-', num * -1
    elif num > 0:
        if rank:
            op, value = '↓', num
        elif pp:
            op, value = '↑', num
        else:
            op, value = '+', num
    else:
        op, value = '', 0
    return [op, value]


def draw_acc(img: Image, acc: float, mode: str):
    acc *= 100
    size = [acc, 100 - acc]
    if mode == 'osu':
        insize = [60, 20, 7, 7, 5, 1]
    elif mode == 'taiko':
        insize = [60, 20, 5, 5, 4, 1]
    elif mode == 'fruits':
        insize = [85, 5, 4, 4, 1, 1]
    else:
        insize = [70, 10, 10, 5, 4, 1]
    insizecolor = ['#ff5858', '#ea7948', '#d99d03', '#72c904', '#0096a2', '#be0089']
    fig = Figure()
    ax = fig.add_axes((0.1, 0.1, 0.8, 0.8))
    patches = ax.pie(size, radius=1, startangle=90, counterclock=False, pctdistance=0.9, wedgeprops=dict(width=0.20), colors=['#66cbfd'])
    ax.pie(insize, radius=0.8, colors=insizecolor, startangle=90, counterclock=False, pctdistance=0.9, wedgeprops=dict(width=0.05))
    patches[0][1].set_alpha(0)
    acc_img = BytesIO()
    fig.savefig(acc_img, transparent=True)
    ax.cla()
    ax.clear()
    fig.clf()
    fig.clear()
    score_acc_img = Image.open(acc_img).convert('RGBA').resize((576, 432))
    img.alpha_composite(score_acc_img, (25, 83))
    return img


async def crop_bg(size: str, path: Union[str, Path]):
    try:
        bg = Image.open(path).convert('RGBA')
    except UnidentifiedImageError:
        os.remove(path)
        data = await get_seasonal_bg()
        pic = SeasonalBackgrounds(**data)
        url = random.choice(pic.backgrounds).url
        res = await safe_async_get(url)
        bg = Image.open(BytesIO(res.content)).convert('RGBA')
    except FileNotFoundError:
        data = await get_seasonal_bg()
        pic = SeasonalBackgrounds(**data)
        url = random.choice(pic.backgrounds).url
        res = await safe_async_get(url)
        bg = Image.open(BytesIO(res.content)).convert('RGBA')
    bg_w, bg_h = bg.size[0], bg.size[1]
    if size == 'BG':
        fix_w = 1500
        fix_h = 720
    elif size == 'H':
        fix_w = 540
        fix_h = 180
    elif size == 'HI':
        fix_w = 1000
        fix_h = 400
    elif size == 'MP':
        fix_w = 1200
        fix_h = 300
    elif size == 'MB':
        fix_w = 1200
        fix_h = 600
    else:
        raise
    # 固定比例
    fix_scale = fix_h / fix_w
    # 图片比例
    bg_scale = bg_h / bg_w
    # 当图片比例大于固定比例
    if bg_scale > fix_scale:
        # 长比例
        scale_width = fix_w / bg_w
        # 等比例缩放
        width = int(scale_width * bg_w)
        height = int(scale_width * bg_h)
        sf = bg.resize((width, height))
        # 计算上下裁切
        crop_height = (height - fix_h) / 2
        x1, y1, x2, y2 = 0, crop_height, width, height - crop_height
        # 裁切保存
        crop_img = sf.crop((x1, y1, x2, y2))
        return crop_img
    # 当图片比例小于固定比例
    elif bg_scale < fix_scale:
        # 宽比例
        scale_height = fix_h / bg_h
        # 等比例缩放
        width = int(scale_height * bg_w)
        height = int(scale_height * bg_h)
        sf = bg.resize((width, height))
        # 计算左右裁切
        crop_width = (width - fix_w) / 2
        x1, y1, x2, y2 = crop_width, 0, width - crop_width, height
        # 裁切保存
        crop_img = sf.crop((x1, y1, x2, y2))
        return crop_img
    else:
        sf = bg.resize((fix_w, fix_h))
        return sf


def stars_diff(mode: Union[str, int], stars: float):
    if mode == 0:
        mode = 'std'
    elif mode == 1:
        mode = 'taiko'
    elif mode == 2:
        mode = 'ctb'
    elif mode == 3:
        mode = 'mania'
    else:
        mode = 'stars'
    default = 115
    if stars < 1:
        xp = 0
        default = 120
    elif stars < 2:
        xp = 120
        default = 120
    elif stars < 3:
        xp = 240
    elif stars < 4:
        xp = 355
    elif stars < 5:
        xp = 470
    elif stars < 6:
        xp = 585
    elif stars < 7:
        xp = 700
    elif stars < 8:
        xp = 815
    else:
        return Image.open(osufile / 'work' / f'{mode}_expertplus.png').convert('RGBA')
    # 取色
    x = (stars - math.floor(stars)) * default + xp
    r, g, b = ColorPic[x, 1]
    # 打开底图
    im = Image.open(osufile / 'work' / f'{mode}.png').convert('RGBA')
    xx, yy = im.size
    # 填充背景
    sm = Image.new('RGBA', im.size, (r, g, b))
    sm.paste(im, (0, 0, xx, yy), im)
    # 把白色变透明
    arr = np.array(sm)
    # 创建mask，将白色替换为True，其他颜色替换为False
    mask = (arr[:, :, 0] == 255) & (arr[:, :, 1] == 255) & (arr[:, :, 2] == 255)
    # 将mask中为True的像素点的alpha通道设置为0
    arr[:, :, 3][mask] = 0
    # 将numpy数组转换回PIL图片
    img = Image.fromarray(arr)
    return img


def get_modeimage(mode: int) -> Path:
    if mode == 0:
        img = 'pfm_std.png'
    elif mode == 1:
        img = 'pfm_taiko.png'
    elif mode == 2:
        img = 'pfm_ctb.png'
    else:
        img = 'pfm_mania.png'
    return osufile / img


def calc_songlen(length: int) -> str:
    map_len = list(divmod(int(length), 60))
    map_len[1] = map_len[1] if map_len[1] >= 10 else f'0{map_len[1]}'
    music_len = f'{map_len[0]}:{map_len[1]}'
    return music_len


async def open_user_icon(info: User) -> Image:
    path = user_cache_path / str(info.id)
    png_user_icon = user_cache_path / str(info.id) / 'icon.png'
    gif_user_icon = user_cache_path / str(info.id) / 'icon.gif'
    # 判断文件是否存在，并读取图片
    if png_user_icon.exists():
        image = Image.open(png_user_icon)
    elif gif_user_icon.exists():
        image = Image.open(gif_user_icon)
    else:
        user_icon = await get_projectimg(info.avatar_url)
        with open(path / f'icon.{info.avatar_url[-3:]}', 'wb') as f:
            f.write(user_icon.getvalue())
        image = Image.open(user_icon)
    return image



from datetime import datetime, timedelta
from typing import Union

from PIL import ImageEnhance, ImageDraw
from nonebot.adapters.onebot.v11 import MessageSegment

from ..api import osu_api, get_map_bg
from ..schema import Beatmap
from ..mods import calc_mods
from ..file import re_map, get_projectimg, download_osu, map_path
from ..utils import FGM
from ..pp import get_ss_pp
from .utils import calc_songlen, draw_fillet, stars_diff, crop_bg, image2bytesio
from .static import *


async def draw_map_info(mapid: int, mods: list) -> Union[str, MessageSegment]:
    info = await osu_api('map', map_id=mapid)
    if not info:
        return '未查询到该地图信息'
    if isinstance(info, str):
        return info
    mapinfo = Beatmap(**info)
    diffinfo = calc_songlen(mapinfo.total_length), mapinfo.bpm, mapinfo.count_circles, mapinfo.count_sliders
    # 获取地图
    path = map_path / str(mapinfo.beatmapset_id)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    osu = path / f"{mapid}.osu"
    if not osu.exists():
        await download_osu(mapinfo.beatmapset_id, mapid)
    ss_pp_info = get_ss_pp(str(osu.absolute()), calc_mods(mods))
    # 计算时间
    if mapinfo.beatmapset.ranked_date:
        old_time = datetime.strptime(mapinfo.beatmapset.ranked_date.replace('Z', ''), '%Y-%m-%dT%H:%M:%S')
        new_time = (old_time + timedelta(hours=8)).strftime('%Y-%m-%d %H:%M:%S')
    else:
        new_time = '谱面状态可能非ranked'
    # BG做地图
    im = Image.new('RGBA', (1200, 600))
    draw = ImageDraw.Draw(im)
    cover = re_map(osu)
    cover_path = path / cover
    if not cover_path.exists():
        bg = await get_map_bg(mapinfo.beatmapset_id, cover)
        with open(cover_path, 'wb') as f:
            f.write(bg.getvalue())
    cover_crop = await crop_bg('MB', cover_path)
    cover_img = ImageEnhance.Brightness(cover_crop).enhance(2 / 4.0)
    im.alpha_composite(cover_img)
    # 获取地图info
    im.alpha_composite(MapBg)
    # 模式
    mode_bg = stars_diff(FGM[mapinfo.mode], ss_pp_info.difficulty.stars)
    mode_img = mode_bg.resize((50, 50))
    im.alpha_composite(mode_img, (50, 100))
    # cs - diff
    mapdiff = [mapinfo.cs, mapinfo.drain, mapinfo.accuracy, mapinfo.ar, ss_pp_info.difficulty.stars]
    for num, i in enumerate(mapdiff):
        color = (255, 255, 255, 255)
        if num == 4:
            color = (255, 204, 34, 255)
        difflen = int(250 * i / 10) if i <= 10 else 250
        diff_len = Image.new('RGBA', (difflen, 8), color)
        im.alpha_composite(diff_len, (890, 426 + 35 * num))
        draw.text((1170, 426 + 35 * num), "%.1f" % i, font=Torus_SemiBold_20, anchor='mm')
    # mapper
    icon_url = f'https://a.ppy.sh/{mapinfo.user_id}'
    user_icon = await get_projectimg(icon_url)
    icon = Image.open(user_icon).convert('RGBA').resize((100, 100))
    icon_img = draw_fillet(icon, 10)
    im.alpha_composite(icon_img, (50, 400))
    # mapid
    draw.text((800, 40), f'Setid: {mapinfo.beatmapset_id}  |  Mapid: {mapid}', font=Torus_Regular_20, anchor='lm')
    # 版本
    draw.text((120, 125), mapinfo.version, font=Torus_SemiBold_25, anchor='lm')
    # 曲名
    draw.text((50, 170), mapinfo.beatmapset.title, font=Torus_SemiBold_30, anchor='lt')
    # 曲师
    draw.text((50, 210), f'by {mapinfo.beatmapset.artist_unicode}', font=Torus_SemiBold_25, anchor='lt')
    # 来源
    draw.text((50, 260), f'Source:{mapinfo.beatmapset.source}', font=Torus_SemiBold_25, anchor='lt')
    # mapper
    draw.text((160, 400), '谱师:', font=Torus_SemiBold_20, anchor='lt')
    draw.text((160, 425), mapinfo.beatmapset.creator, font=Torus_SemiBold_20, anchor='lt')
    # ranked时间
    draw.text((160, 460), '上架时间:', font=Torus_SemiBold_20, anchor='lt')
    draw.text((160, 485), new_time, font=Torus_SemiBold_20, anchor='lt')
    # 状态
    draw.text((1100, 304), mapinfo.status.capitalize(), font=Torus_SemiBold_20, anchor='mm')
    # 时长 - 滑条
    for num, i in enumerate(diffinfo):
        draw.text((770 + 120 * num, 365), f'{i}', font=Torus_Regular_20, anchor='lm', fill=(255, 204, 34, 255))
    # maxcb
    draw.text((50, 570), f'最大连击: {mapinfo.max_combo}', font=Torus_SemiBold_20, anchor='lm')
    # pp
    draw.text((320, 570), f'SS PP: {int(round(ss_pp_info.pp, 0))}', font=Torus_SemiBold_20, anchor='lm')
    # 输出
    base = image2bytesio(im)
    im.close()
    msg = MessageSegment.image(base)
    return msg



from pathlib import Path
from PIL import Image, ImageFont

osufile = Path(__file__).parent.parent / 'osufile'
Torus_Regular_20 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus Regular.otf'), 20)
Torus_Regular_25 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus Regular.otf'), 25)
Torus_Regular_30 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus Regular.otf'), 30)
Torus_Regular_35 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus Regular.otf'), 35)
Torus_Regular_40 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus Regular.otf'), 40)
Torus_Regular_45 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus Regular.otf'), 45)
Torus_Regular_50 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus Regular.otf'), 50)
Torus_Regular_75 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus Regular.otf'), 75)
Torus_SemiBold_20 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus SemiBold.otf'), 20)
Torus_SemiBold_25 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus SemiBold.otf'), 25)
Torus_SemiBold_30 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus SemiBold.otf'), 30)
Torus_SemiBold_40 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus SemiBold.otf'), 40)
Torus_SemiBold_50 = ImageFont.truetype(str(osufile / 'fonts' / 'Torus SemiBold.otf'), 50)
Venera_75 = ImageFont.truetype(str(osufile / 'fonts' / 'Venera.otf'), 75)

ColorPic = Image.open(osufile / 'work' / 'color.png').load()
InfoImg = Image.open(osufile / 'info.png').convert('RGBA')
NewInfoImg = Image.open(osufile / 'info_new.png').convert('RGBA')
SupporterBg = Image.open(osufile / 'work' / 'suppoter.png').convert('RGBA')
ExpLeftBg = Image.open(osufile / 'work' / 'left.png').convert('RGBA')
ExpCenterBg = Image.open(osufile / 'work' / 'center.png').convert('RGBA')
ExpRightBg = Image.open(osufile / 'work' / 'right.png').convert('RGBA')
BgImg = Image.open(osufile / 'Best Performance.png').convert('RGBA')
MapBg = Image.open(osufile / 'beatmapinfo.png').convert('RGBA')
BarImg = Image.open(osufile / 'work' / 'bmap.png').convert('RGBA')



from typing import Literal, List, Optional

from .basemodel import Base


class Badge(Base):
    awarded_at: str
    description: str
    image_url: str
    url: str


class GradeCounts(Base):
    ssh: int
    ss: int
    sh: int
    s: int
    a: int


class Level(Base):
    current: int
    progress: int


class UserStatistics(Base):
    grade_counts: GradeCounts
    hit_accuracy: float
    is_ranked: bool
    level: Level
    maximum_combo: int
    play_count: int
    play_time: int
    pp: int
    ranked_score: int
    replays_watched_by_others: int
    total_hits: int
    total_score: int
    global_rank: Optional[int]
    country_rank: Optional[int]
    badges: Optional[List[Badge]]


class UserCompact(Base):
    avatar_url: str
    country_code: str
    default_group: str
    id: int
    is_active: bool
    is_bot: bool
    is_deleted: bool
    is_online: bool
    is_supporter: bool
    last_visit: Optional[str]
    profile_colour: Optional[str]
    username: str
    statistics: Optional[UserStatistics]


class User(UserCompact):
    cover_url: Optional[str]
    has_supported: Optional[bool]
    join_date: Optional[str]
    location: Optional[str]
    occupation: Optional[str]
    playmode: Optional[Literal['fruits', 'mania', 'osu', 'taiko']]
    playstyle: Optional[List[str]]
    badges: Optional[List[Badge]]




from typing import List, Optional

from .basemodel import Base


class ListData(Base):
    accurate: bool
    currentMod: Optional[List[str]]
    currentPP: Optional[int]
    currentScore: Optional[int]
    currentScoreLink: Optional[str]
    difficulty: float
    id: str
    mapCoverUrl: str
    mapLink: str
    mapName: str
    mod: List[str]
    newRecordPercent: float
    passPercent: float
    ppIncrement: float
    ppIncrementExpect: float
    predictPP: float


class Data(Base):
    next: int
    prev: int
    total: int
    list: List[ListData]


class RecommendData(Base):
    code: int
    message: str
    success: bool
    data: Optional[Data]



from typing import List, Literal, Optional

from .basemodel import Base
from .beatmap import Beatmap, Beatmapset


class Statistics(Base):
    count_50: int
    count_100: int
    count_300: int
    count_geki: int
    count_katu: int
    count_miss: int


class Score(Base):
    id: int
    best_id: Optional[int]
    user_id: int
    accuracy: float
    mods: List[str]
    score: int
    max_combo: int
    perfect: int
    statistics: Statistics
    passed: bool
    pp: Optional[float]
    rank: str
    created_at: str
    mode: Literal['fruits', 'mania', 'osu', 'taiko']
    mode_int: int
    beatmap: Optional[Beatmap]
    beatmapset: Optional[Beatmapset]


class BeatmapUserScore(Base):
    position: int
    score: Score



from typing import Union
from nonebot.internal.params import Depends
from nonebot.adapters.onebot.v11 import Message, MessageEvent
from nonebot_plugin_guild_patch import GuildMessageEvent
from nonebot.params import T_State, CommandArg
from ..utils import mods2list
from ..database.models import UserData


def split_msg():
    async def dependency(event: Union[MessageEvent, GuildMessageEvent], state: T_State, msg: Message = CommandArg()):
        qq = event.user_id
        for msg_seg in event.message:
            if msg_seg.type == "at":
                qq = str(msg_seg.data.get("qq", ""))
        user_data = await UserData.get_or_none(user_id=qq)
        state['user'] = user_data.osu_id if user_data else 0
        state['mode'] = str(user_data.osu_mode) if user_data else '0'
        state['mods'] = []
        state['day'] = 0
        state['is_name'] = False
        symbol_ls = [':', '+', '：', '#', '＃']
        symbol_dic = {':': 'mode', '+': 'mods', '：': 'mode', '#': 'day', '＃': 'day'}
        dic = {}
        arg = msg.extract_plain_text().strip()
        if max([arg.find(i) for i in symbol_ls]) >= 0:
            for i in symbol_ls:
                dic[i] = arg.find(i)
            sorted_dict = sorted(dic.items(), key=lambda x: x[1])
            for i in range(len(sorted_dict) - 1):
                if sorted_dict[i][1] >= 0:
                    state[symbol_dic[sorted_dict[i][0]]] = arg[sorted_dict[i][1] + 1:sorted_dict[i + 1][1]].strip()
            if sorted_dict[-1][1] >= 0:
                state[symbol_dic[sorted_dict[-1][0]]] = arg[sorted_dict[-1][1] + 1:].strip()
            if isinstance(state['mods'], str):
                state['mods'] = mods2list(state['mods'].strip())
            index = min([arg.find(i) for i in symbol_ls if arg.find(i) >= 0])
            state['para'] = arg[:index].strip()
        else:
            state['para'] = arg.strip()
        if state['_prefix']['command'][0] in ('pr', 're', 'info', 'tbp', 'recent') and state['para']:
            state['is_name'] = True
        # 分出user和参数
        if state['para'].find(' ') > 0 and state['_prefix']['command'][0] not in ('pr', 're', 'info', 'tbp', 'recent'):
            state['user'] = state['para'][:state['para'].rfind(' ')].strip()
            state['para'] = state['para'][state['para'].rfind(' ') + 1:].strip()
            state['is_name'] = True
        elif state['para'].find(' ') > 0 and state['_prefix']['command'][0] in ('pr', 're', 'info', 'tbp', 'recent'):
            state['user'] = state['para']
        if not state['mode'].isdigit() and (int(state['mode']) < 0 or int(state['mode']) > 3):
            state['error'] = '模式应为0-3的数字！'
        if isinstance(state['day'], str) and not state['day'].isdigit() and (int(state['day']) < 0):
            state['error'] = '查询的日期应是一个正数'
        if state['user'] == 0 and not state['para']:
            state['error'] = '该账号尚未绑定，请输入 /bind 用户名 绑定账号'
        state['day'] = int(state['day'])
    return Depends(dependency)



import re
import json
from typing import List, Tuple
from pathlib import Path

from nonebot import on_command
from nonebot.adapters.onebot.v11 import Message, MessageSegment, Bot, MessageEvent, GroupMessageEvent
from nonebot.params import CommandArg
from ..api import safe_async_get


medal_data_path = Path(__file__).parent.parent / 'osufile' / 'medals' / 'medals.json'
with open(medal_data_path, 'r') as file:
    medal_json = json.load(file)


async def send_forward_msg(
    bot: Bot,
    event: MessageEvent,
    user_message: List[Tuple[str, str, Message]],
):
    """
    发送 forward 消息

    > 参数：
        - bot: Bot 对象
        - event: MessageEvent 对象
        - user_message: 合并消息的用户信息列表

    > 返回值：
        - 成功：返回消息发送结果
        - 失败：抛出异常
    """

    def to_json(info: Tuple[str, str, Message]):
        """
        将消息转换为 forward 消息的 json 格式
        """
        return {
            "type": "node",
            "data": {"name": info[0], "uin": info[1], "content": info[2]},
        }

    messages = [to_json(info) for info in user_message]
    if isinstance(event, GroupMessageEvent):
        return await bot.call_api(
            "send_group_forward_msg", group_id=event.group_id, messages=messages
        )
    else:
        return await bot.call_api(
            "send_private_forward_msg", user_id=event.user_id, messages=messages
        )


medal = on_command('medal', aliases={'成就'}, priority=11, block=True)


@medal.handle()
async def _(bot: Bot, event: MessageEvent, msg: Message = CommandArg()):
    name = msg.extract_plain_text()
    data = await safe_async_get(f'https://osekai.net/medals/api/public/get_medal.php?medal={name}')
    medal_data = data.json()
    if 'MedalID' not in medal_data:
        await medal.finish('没有找到欸，看看是不是名字打错了')
    words = ''
    if medal_data['Restriction'] != 'NULL':
        words += f"限制模式：{medal_data['Restriction']}\n"
    words += '获得方式：\n'
    if medal_data['Name'] in medal_json:
        words += medal_json[medal_data['Name']]['MedalSolution']
    else:
        words += medal_data['Solution'] if medal_data['Solution'] else medal_data['Instructions']
        table_regex = r"<table[^>]*>(.*?)<\/table>"
        table_match = re.search(table_regex, words, re.DOTALL)
        if table_match:
            table_text = table_match.group(1)

            # 使用正则表达式匹配并提取表格行部分
            row_regex = r"<tr[^>]*>(.*?)<\/tr>"
            rows = re.findall(row_regex, table_text, re.DOTALL)

            result = ""

            # 遍历表格行并提取单元格内容
            for row in rows:
                # 使用正则表达式匹配并提取单元格部分
                cell_regex = r"<t[hd][^>]*>(.*?)<\/t[hd]>"
                cells = re.findall(cell_regex, row, re.DOTALL)
                for cell in cells:
                    # 去除单元格内的HTML标签
                    cell_text = re.sub(r"<[^>]*>", "", cell)
                    result += cell_text + " "
                result += "\n"

            # 将提取的表格文字替换回原文中
            words = re.sub(table_regex, result, words)
    style_regex = r"<style[^>]*>(.*?)<\/style>"
    words = re.sub(style_regex, "", words)
    words = re.sub(r"<[^>]+>", "", words)
    if medal_data['PackID']:
        words += f'\nhttps://osu.ppy.sh/beatmaps/packs/{medal_data["PackID"].rstrip(",,,")}'
    await medal.send(MessageSegment.reply(event.message_id) + MessageSegment.image(medal_data['Link']) + words)
    if medal_data['beatmaps']:
        msg_ls = []
        for beatmap in medal_data['beatmaps']:
            msg = MessageSegment.image(f'https://assets.ppy.sh/beatmaps/{beatmap["MapsetID"]}/covers/cover.jpg') +\
                  f'{beatmap["SongTitle"]} [{beatmap["DifficultyName"]}]\n{beatmap["Difficulty"]}⭐\n' + \
                  f'https://osu.ppy.sh/b/{beatmap["BeatmapID"]}'
            msg_ls.append(('推荐谱面', bot.self_id, msg))
        await send_forward_msg(bot, event, msg_ls)



from pydantic import Extra, BaseSettings
from typing import Optional, List, Union


class Config(BaseSettings):
    osu_client: Optional[int] = None
    osu_key: Optional[str] = None
    osz_mirror: Optional[int] = 0
    info_bg: Optional[List[str]] = ['https://api.gumengya.com/Api/DmImg?format=image']
    osu_proxy: Optional[Union[str, dict]] = None

    class Config:
        extra = Extra.ignore
        case_sensitive = False



from typing import List, Optional

from .basemodel import Base


class BidData(Base):
    bid: int
    mode: int
    version: str
    length: int
    CS: float
    AR: float
    OD: float
    HP: float
    star: float
    aim: float
    speed: float
    hit300window: float
    pp: float
    pp_aim: float
    pp_speed: float
    pp_acc: float
    circles: int
    sliders: int
    spinners: int
    maxcombo: int
    playcount: int
    passcount: int
    bg: str
    audio: str


class MapInfo(Base):
    sid: int
    local_update: int
    bids_amount: int
    approved: int
    title: str
    artist: str
    titleU: str
    artistU: str
    creator: str
    creator_id: int
    source: str
    last_update: str
    approved_date: int
    bpm: float
    favourite_count: int
    video: int
    storyboard: int
    tags: str
    language: int
    genre: str
    bid_data: List[BidData]


class SayoBeatmap(Base):
    status: int
    data: Optional[MapInfo]



from typing import List
from .schema import Score

mods_dic = {
    'NO': 0,
    'NF': 1 << 0,
    'EZ': 1 << 1,
    'TD': 1 << 2,
    'HD': 1 << 3,
    'HR': 1 << 4,
    'SD': 1 << 5,
    'DT': 1 << 6,
    'RX': 1 << 7,
    'HT': 1 << 8,
    'NC': 1 << 9,
    'FL': 1 << 10,
    'AT': 1 << 11,
    'SO': 1 << 12,
    'RX2': 1 << 13,
    'PF': 1 << 14,
    '4K': 1 << 15,
    '5K': 1 << 16,
    '6K': 1 << 17,
    '7K': 1 << 18,
    '8K': 1 << 19,
    'FI': 1 << 20,
    'RD': 1 << 21,
    'Cinema': 1 << 22,
    'TG': 1 << 23,
    '9K': 1 << 24,
    'KC': 1 << 25,
    '1K': 1 << 26,
    '3K': 1 << 27,
    '2K': 1 << 28,
    'V2': 1 << 29,
    'MR': 1 << 30
}


def get_mods_list(score_ls: List[Score], mods: List[str]) -> List[int]:
    if not mods:
        return list(range(len(score_ls)))
    mods_index_ls = []
    for i, score in enumerate(score_ls):
        if score.mods and set(mods).issubset(set(score.mods)):
            mods_index_ls.append(i)
    return mods_index_ls


def calc_mods(mods: List[str]) -> int:
    num = 0
    for mod in mods:
        num ^= mods_dic[mod.upper()]
    return num



from rosu_pp_py import Beatmap, Calculator, PerformanceAttributes
from .schema import Score
from .mods import calc_mods


def cal_pp(score: Score, path: str) -> PerformanceAttributes:
    beatmap = Beatmap(path=path)
    mods = calc_mods(score.mods)
    if mods & (1 << 9):
        mods -= 1 << 9
        mods += 1 << 6
    if score.mode_int == 2:
        c = Calculator(acc=score.accuracy * 100, n_katu=score.statistics.count_katu,
                       combo=score.max_combo,
                       n_misses=score.statistics.count_miss,
                       n100=score.statistics.count_100,
                       n300=score.statistics.count_300,
                       mods=mods, mode=score.mode_int)
    else:
        c = Calculator(acc=score.accuracy * 100, n_katu=score.statistics.count_katu,
                       n_geki=score.statistics.count_geki, combo=score.max_combo,
                       n_misses=score.statistics.count_miss,
                       n50=score.statistics.count_50,
                       n100=score.statistics.count_100,
                       n300=score.statistics.count_300,
                       mods=mods, mode=score.mode_int)
    return c.performance(beatmap)


def get_if_pp_ss_pp(score: Score, path: str) -> tuple:
    beatmap = Beatmap(path=path)
    mods = calc_mods(score.mods)
    if mods & (1 << 9):
        mods -= 1 << 9
        mods += 1 << 6
    c = Calculator(acc=score.accuracy * 100, n_katu=score.statistics.count_katu,
                   n_geki=score.statistics.count_geki,
                   n50=score.statistics.count_50,
                   n100=score.statistics.count_100,
                   n300=score.statistics.count_300 + score.statistics.count_miss,
                   mods=mods, mode=score.mode_int)
    if_pp = c.performance(beatmap).pp
    c = Calculator(acc=100,
                   mods=mods, mode=score.mode_int)
    ss_pp = c.performance(beatmap).pp
    return str(int(round(if_pp, 0))), str(int(round(ss_pp, 0)))


def get_ss_pp(path: str, mods: int) -> PerformanceAttributes:
    beatmap = Beatmap(path=path)
    if mods & (1 << 9):
        mods -= 1 << 9
        mods += 1 << 6
    c = Calculator(acc=100, mods=mods)
    ss_pp_info = c.performance(beatmap)
    return ss_pp_info



import re
import shutil
import urllib
import asyncio
from dataclasses import dataclass
from pathlib import Path
from random import shuffle
from typing import List, Union
from expiringdict import ExpiringDict

from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent, Message, MessageEvent, MessageSegment, ActionFailed
from nonebot.adapters.onebot.v11.helpers import ImageURLs
from nonebot.permission import SUPERUSER
from nonebot_plugin_guild_patch import GuildMessageEvent
from nonebot.exception import ParserExit
from nonebot.internal.params import Depends
from nonebot.params import T_State, ShellCommandArgv, CommandArg, RegexGroup
from nonebot.plugin import PluginMetadata
from nonebot.rule import ArgumentParser
from nonebot.log import logger
from nonebot import on_command, require, on_shell_command, on_regex, get_driver
from nonebot_plugin_tortoise_orm import add_model
from .draw import draw_info, draw_score, draw_map_info, draw_bmap_info, draw_bp, image2bytesio, get_score_data
from .file import download_map, download_osu, download_tmp_osu, user_cache_path, save_info_pic
from .utils import NGM, GMN, mods2list
from .database.models import UserData
from .mania import generate_preview_pic, convert_mania_map, Options
from .api import osu_api, get_sayo_map_info, get_recommend, update_recommend
from .info import get_bg, bind_user_info, update_user_info
from .matcher import *
from .config import Config


require('nonebot_plugin_apscheduler')
from nonebot_plugin_apscheduler import scheduler


@dataclass
class ReviewData:
    msg_id: int
    pic_url: str
    group: int
    user: str
    id: int


review_pic_ls: List[ReviewData] = []
counter = 0
recommend_cache = ExpiringDict(1000, 60 * 60 * 12)
usage = "发送/osuhelp 查看帮助"
detail_usage = """以下<>内是必填内容，()内是选填内容，user可以是用户名也可以@他人，mode为0-3的一个数字
/info (user)(:mode)
/re (user)(:mode)
/score <mapid>(:mode)(+mods)
/bp (user) <num> (:mode)(+mods)
/pfm (user) <min>-<max> (:mode)(+mods)
/tbp (user) (:mode)
/map <mapid> (+mods)
/倍速 <setid> (rate)-(rate)
/反键 <mapid> (gap) (ln_as_hit_thres)
其中gap为ln的间距默认为150 (ms)
ln_as_hit_thres为ln转为note的阈值默认为100 (ms)
rate可为任意小数"""

__plugin_meta__ = PluginMetadata(
    name="OSUBot",
    description="OSU查分插件",
    usage=usage,
    type='application',
    homepage='https://github.com/yaowan233/nonebot-plugin-osubot',
    config=Config,
    supported_adapters={"~onebot.v11", "~qqguild"},
    extra={
        "unique_name": "osubot",
        "author": "yaowan233 <572473053@qq.com>",
        "version": "3.4.1",
    },
)

add_model('nonebot_plugin_osubot.database.models')


parser = ArgumentParser('convert', description='变换mania谱面')
parser.add_argument('--set', type=int, help='要转换的谱面的setid')
parser.add_argument('--map', type=int, help='要转换的谱面的mapid')
parser.add_argument('--fln', action='store_true', help='将谱面转换为反键')
parser.add_argument('--rate', type=float, help='谱面倍速速率')
parser.add_argument('--end_rate', type=float, help='谱面倍速速率的最大值')
parser.add_argument('--step', type=float, help='谱面倍速的step')
parser.add_argument('--od', type=float, help='改变谱面od到指定值')
parser.add_argument('--nsv', action='store_true', help='移除谱面所有sv')
parser.add_argument('--nln', action='store_true', help='移除谱面所有ln')
parser.add_argument('--gap', nargs='?', default='150', type=float, help='指定反键的间距时间，默认150ms')
parser.add_argument('--thres', nargs='?', default='100', type=float, help='指定转反键时ln转换为note的阈值，默认100ms')

convert = on_shell_command("convert", parser=parser, block=True, priority=13)


@convert.handle()
async def _(
        bot: Bot, event: Union[GroupMessageEvent, GuildMessageEvent], argv: List[str] = ShellCommandArgv()
):
    if isinstance(event, GuildMessageEvent):
        await convert.finish(MessageSegment.reply(event.message_id) + '很抱歉，频道暂不支持上传文件')
    try:
        args = parser.parse_args(argv)
    except ParserExit as e:
        if e.status == 0:
            await convert.finish(MessageSegment.reply(event.message_id) + MessageSegment.image(
                Path(__file__).parent / 'osufile' / 'convert.jpg'))
        await convert.finish(MessageSegment.reply(event.message_id) + str(e))
        return
    options = Options(**vars(args))
    if options.map:
        sayo_map_info = await get_sayo_map_info(options.map, 1)
        options.set = sayo_map_info.data.sid
        options.sayo_info = sayo_map_info
    if not options.set:
        await convert.finish(MessageSegment.reply(event.message_id) + '请提供需要转换的谱面setid')
    if options.nln and options.fln:
        await convert.finish(MessageSegment.reply(event.message_id) + '指令矛盾！')
    osz_file = await convert_mania_map(options)
    if not osz_file:
        await convert.finish(MessageSegment.reply(event.message_id) + '未找到该地图，请检查是否搞混了mapID与setID')
    name = urllib.parse.unquote(osz_file.name)
    file_path = osz_file.absolute()
    try:
        await bot.upload_group_file(group_id=event.group_id, file=str(file_path), name=name)
    except ActionFailed:
        await convert.finish(MessageSegment.reply(event.message_id) + '上传文件失败，可能是群空间满或没有权限导致的')
    finally:
        try:
            osz_file.unlink()
        except PermissionError:
            ...

info = on_command("info", aliases={'Info', 'INFO'}, block=True, priority=11)


@info.handle(parameterless=[split_msg()])
async def _info(state: T_State, event: Union[MessageEvent, GuildMessageEvent]):
    if 'error' in state:
        await info.finish(MessageSegment.reply(event.message_id) + state['error'])
    user = state['para'] if state['para'] else state['user']
    mode = state['mode']
    day = state['day']
    data = await draw_info(user, NGM[mode], day, state['is_name'])
    await info.finish(MessageSegment.reply(event.message_id) + data)


mu = on_command("mu", aliases={'Mu', 'MU'}, block=True, priority=11)


@mu.handle(parameterless=[split_msg()])
async def _mu(state: T_State, event: Union[MessageEvent, GuildMessageEvent]):
    if 'error' in state:
        await info.finish(MessageSegment.reply(event.message_id) + state['error'])
    user_id = state['user']
    data = f"https://osu.ppy.sh/u/{user_id}"
    await mu.finish(MessageSegment.reply(event.message_id) + data)


recent = on_command("recent", aliases={'re', 'RE', 'Re'}, priority=11, block=True)


@recent.handle(parameterless=[split_msg()])
async def _recent(state: T_State, event: Union[MessageEvent, GuildMessageEvent]):
    if 'error' in state:
        await recent.finish(MessageSegment.reply(event.message_id) + state['error'])
    user = state['para'] if state['para'] else state['user']
    mode = state['mode']
    if state['day'] == 0:
        state['day'] = 1
    data = await draw_score('recent', user, NGM[mode], [], state['day'] - 1, is_name=state['is_name'])
    await recent.finish(MessageSegment.reply(event.message_id) + data)

pr = on_command("pr", priority=11, block=True, aliases={'PR', 'Pr'})


@pr.handle(parameterless=[split_msg()])
async def _pr(state: T_State, event: Union[MessageEvent, GuildMessageEvent]):
    if 'error' in state:
        await pr.finish(MessageSegment.reply(event.message_id) + state['error'])
    user = state['para'] if state['para'] else state['user']
    mode = state['mode']
    if state['day'] == 0:
        state['day'] = 1
    data = await draw_score('pr', user, NGM[mode], [], state['day'] - 1, is_name=state['is_name'])
    await pr.finish(MessageSegment.reply(event.message_id) + data)

score = on_command('score', priority=11, block=True)


@score.handle(parameterless=[split_msg()])
async def _score(state: T_State, event: Union[MessageEvent, GuildMessageEvent]):
    if 'error' in state:
        await score.finish(MessageSegment.reply(event.message_id) + state['error'])
    user = state['user']
    mode = state['mode']
    mods = state['mods']
    map_id = state['para']
    data = await get_score_data(user, NGM[mode], mapid=map_id, mods=mods, is_name=state['is_name'])
    await score.finish(MessageSegment.reply(event.message_id) + data)


bp = on_command('bp', priority=11, block=True)


@bp.handle(parameterless=[split_msg()])
async def _bp(state: T_State, event: Union[MessageEvent, GuildMessageEvent]):
    if 'error' in state:
        await bp.finish(MessageSegment.reply(event.message_id) + state['error'])
    user = state['user']
    mode = state['mode']
    mods = state['mods']
    para = state['para']
    if '-' in para:
        await _pfm(state, event)
        return
    if not para.isdigit():
        await bp.finish(MessageSegment.reply(event.message_id) + '只能接受纯数字的bp参数')
    best = int(para)
    if best <= 0 or best > 100:
        await bp.finish(MessageSegment.reply(event.message_id) + '只允许查询bp 1-100 的成绩')
    data = await draw_score('bp', user, NGM[mode], best=best, mods=mods, is_name=state['is_name'])
    await bp.finish(MessageSegment.reply(event.message_id) + data)


pfm = on_command('pfm', priority=11, block=True)


@pfm.handle(parameterless=[split_msg()])
async def _pfm(state: T_State, event: Union[MessageEvent, GuildMessageEvent]):
    if 'error' in state:
        await pfm.finish(MessageSegment.reply(event.message_id) + state['error'])
    user = state['user']
    mode = state['mode']
    mods = state['mods']
    para = state['para']
    ls = para.split('-')
    low, high = ls[0], ls[1]
    if not low.isdigit() or not high.isdigit():
        await pfm.finish(MessageSegment.reply(event.message_id) + '参数应为 "数字-数字"的形式!')
        return
    low, high = int(low), int(high)
    if not 0 < low < high <= 100:
        await pfm.finish(MessageSegment.reply(event.message_id) + '仅支持查询bp1-100')
    data = await draw_bp('bp', user, NGM[mode], mods, low, high, is_name=state['is_name'])
    await pfm.finish(MessageSegment.reply(event.message_id) + data)


tbp = on_command('tbp', aliases={'todaybp'}, priority=11, block=True)


@tbp.handle(parameterless=[split_msg()])
async def _tbp(state: T_State, event: Union[MessageEvent, GuildMessageEvent]):
    if 'error' in state:
        await tbp.finish(MessageSegment.reply(event.message_id) + state['error'])
    user = state['para'] if state['para'] else state['user']
    mode = state['mode']
    day = state['day']
    data = await draw_bp('tbp', user, NGM[mode], [], day=day, is_name=state['is_name'])
    await tbp.finish(MessageSegment.reply(event.message_id) + data)


osu_map = on_command('map', priority=11, block=True)


@osu_map.handle(parameterless=[split_msg()])
async def _map(state: T_State, event: Union[MessageEvent, GuildMessageEvent]):
    map_id = state['para']
    mods = state['mods']
    if not map_id:
        await osu_map.finish(MessageSegment.reply(event.message_id) + '请输入地图ID')
    elif not map_id.isdigit():
        await osu_map.finish(MessageSegment.reply(event.message_id) + '请输入正确的地图ID')
    m = await draw_map_info(map_id, mods)
    await osu_map.finish(MessageSegment.reply(event.message_id) + m)


bmap = on_command('bmap', priority=11, block=True)


@bmap.handle(parameterless=[split_msg()])
async def _bmap(state: T_State, event: Union[MessageEvent, GuildMessageEvent]):
    set_id = state['para']
    if not set_id:
        await bmap.finish(MessageSegment.reply(event.message_id) + '请输入setID')
    if not set_id.isdigit():
        await bmap.finish(MessageSegment.reply(event.message_id) + '请输入正确的setID')
        return
    m = await draw_bmap_info(set_id)
    await bmap.finish(MessageSegment.reply(event.message_id) + m)


osudl = on_command('osudl', priority=11, block=True)


@osudl.handle()
async def _osudl(bot: Bot, event: Union[GroupMessageEvent, GuildMessageEvent], msg: Message = CommandArg()):
    if isinstance(event, GuildMessageEvent):
        await convert.finish(MessageSegment.reply(event.message_id) + '很抱歉，频道暂不支持上传文件')
    setid = msg.extract_plain_text().strip()
    if not setid:
        return
    if not setid.isdigit():
        await osudl.finish(MessageSegment.reply(event.message_id) + '请输入正确的地图ID')
    osz_path = await download_map(int(setid))
    name = urllib.parse.unquote(osz_path.name)
    file_path = osz_path.absolute()
    try:
        await bot.upload_group_file(group_id=event.group_id, file=str(file_path), name=name)
    except ActionFailed:
        await osudl.finish(MessageSegment.reply(event.message_id) + '上传文件失败，可能是群空间满或没有权限导致的')
    finally:
        try:
            osz_path.unlink()
        except PermissionError:
            ...

bind = on_command('bind', priority=11, block=True)
lock = asyncio.Lock()


@bind.handle()
async def _bind(event: Union[MessageEvent, GuildMessageEvent], msg: Message = CommandArg()):
    name = msg.extract_plain_text()
    if not name:
        await bind.finish(MessageSegment.reply(event.message_id) + '请输入您的 osuid')
    async with lock:
        if _ := await UserData.get_or_none(user_id=event.get_user_id()):
            await bind.finish(MessageSegment.reply(event.message_id) + '您已绑定，如需要解绑请输入/unbind')
        msg = await bind_user_info('bind', name, event.get_user_id(), True)
    await bind.finish(MessageSegment.reply(event.message_id) + msg)


unbind = on_command('unbind', priority=11, block=True)


@unbind.handle()
async def _unbind(event: Union[MessageEvent, GuildMessageEvent]):
    if _ := await UserData.get_or_none(user_id=event.get_user_id()):
        await UserData.filter(user_id=event.get_user_id()).delete()
        await unbind.finish(MessageSegment.reply(event.message_id) + '解绑成功！')
    else:
        await unbind.finish(MessageSegment.reply(event.message_id) + '尚未绑定，无需解绑')


update = on_command('更新模式', priority=11, aliases={'更改模式'}, block=True)


@update.handle()
async def _(event: Union[MessageEvent, GuildMessageEvent], msg: Message = CommandArg()):
    args = msg.extract_plain_text().strip()
    user = await UserData.get_or_none(user_id=event.get_user_id())
    if not user:
        await update.finish(MessageSegment.reply(event.message_id) + '该账号尚未绑定，请输入 /bind 用户名 绑定账号')
    elif not args:
        await update.finish(MessageSegment.reply(event.message_id) + '请输入需要更新内容的模式')
    if not args.isdigit():
        await update.finish(MessageSegment.reply(event.message_id) + '请输入正确的模式 0-3')
        return
    mode = int(args)
    if 0 <= mode < 4:
        await UserData.filter(user_id=event.get_user_id()).update(osu_mode=mode)
        msg = f'已将默认模式更改为 {NGM[str(mode)]}'
    else:
        msg = '请输入正确的模式 0-3'
    await update.finish(MessageSegment.reply(event.message_id) + msg)


getbg = on_command('getbg', priority=11, block=True)


@getbg.handle()
async def _get_bg(event: Union[MessageEvent, GuildMessageEvent], msg: Message = CommandArg()):
    bg_id = msg.extract_plain_text().strip()
    if not bg_id:
        msg = '请输入需要提取BG的地图ID'
    else:
        byt = await get_bg(bg_id)
        if isinstance(msg, str):
            await getbg.finish(MessageSegment.reply(event.message_id) + msg)
        msg = MessageSegment.image(byt)
    await getbg.finish(MessageSegment.reply(event.message_id) + msg)

change = on_command('倍速', priority=11, block=True)


@change.handle()
async def _(bot: Bot, event: Union[GroupMessageEvent, GuildMessageEvent], msg: Message = CommandArg()):
    if isinstance(event, GuildMessageEvent):
        await convert.finish(MessageSegment.reply(event.message_id) + '很抱歉，频道暂不支持上传文件')
    args = msg.extract_plain_text().strip().split()
    argv = ['--map']
    if not args:
        await change.finish(MessageSegment.reply(event.message_id) + '请输入需要倍速的地图mapID')
    set_id = args[0]
    if not set_id.isdigit():
        await change.finish(MessageSegment.reply(event.message_id) + '请输入正确的mapID')
    argv.append(set_id)
    if len(args) >= 2:
        argv.append('--rate')
        if '-' in args[1]:
            low, high = args[1].split('-')
            argv.extend([low, '--end_rate', high, '--step', '0.05'])
        else:
            argv.append(args[1])
    else:
        await change.finish(MessageSegment.reply(event.message_id) + '请输入倍速速率')
    args = parser.parse_args(argv)
    options = Options(**vars(args))
    if options.map:
        sayo_map_info = await get_sayo_map_info(options.map, 1)
        options.set = sayo_map_info.data.sid
        options.sayo_info = sayo_map_info
    osz_path = await convert_mania_map(options)
    if not osz_path:
        await change.finish(MessageSegment.reply(event.message_id) + '未找到该地图，请检查是否搞混了mapID与setID')
    name = urllib.parse.unquote(osz_path.name)
    file_path = osz_path.absolute()
    try:
        await bot.upload_group_file(group_id=event.group_id, file=str(file_path), name=name)
    except ActionFailed:
        await change.finish(MessageSegment.reply(event.message_id) + '上传文件失败，可能是群空间满或没有权限导致的')
    finally:
        try:
            osz_path.unlink()
        except PermissionError:
            ...

generate_full_ln = on_command('反键', priority=11, block=True)


@generate_full_ln.handle()
async def _(bot: Bot, event: Union[GroupMessageEvent, GuildMessageEvent], msg: Message = CommandArg()):
    if isinstance(event, GuildMessageEvent):
        await convert.finish(MessageSegment.reply(event.message_id) + '很抱歉，频道暂不支持上传文件')
    args = msg.extract_plain_text().strip().split()
    if not args:
        await generate_full_ln.finish(MessageSegment.reply(event.message_id) + '请输入需要转ln的地图setID')
    set_id = args[0]
    if not set_id.isdigit():
        await generate_full_ln.finish(MessageSegment.reply(event.message_id) + '请输入正确的setID')
    argv = ['--set', set_id, '--fln']
    if len(args) >= 2:
        argv.append('--gap')
        argv.append(args[1])
    if len(args) >= 3:
        argv.append('--thres')
        argv.append(args[2])
    args = parser.parse_args(argv)
    options = Options(**vars(args))
    osz_path = await convert_mania_map(options)
    if not osz_path:
        await generate_full_ln.finish(MessageSegment.reply(event.message_id) + '未找到该地图，请检查是否搞混了mapID与setID')
    name = urllib.parse.unquote(osz_path.name)
    file_path = osz_path.absolute()
    try:
        await bot.upload_group_file(group_id=event.group_id, file=str(file_path), name=name)
    except ActionFailed:
        await generate_full_ln.finish(MessageSegment.reply(event.message_id) + '上传文件失败，可能是群空间满或没有权限导致的')
    finally:
        try:
            osz_path.unlink()
        except PermissionError:
            ...

generate_preview = on_command('预览', aliases={'preview'}, priority=11, block=True)


@generate_preview.handle()
async def _(event: Union[MessageEvent, GuildMessageEvent], msg: Message = CommandArg()):
    osu_id = msg.extract_plain_text().strip()
    if not osu_id or not osu_id.isdigit():
        await osudl.finish(MessageSegment.reply(event.message_id) + '请输入正确的地图mapID')
    data = await osu_api('map', map_id=int(osu_id))
    if not data:
        await generate_preview.finish(MessageSegment.reply(event.message_id) + '未查询到该地图')
    if isinstance(data, str):
        await generate_preview.finish(MessageSegment.reply(event.message_id) + data)
    osu = await download_tmp_osu(osu_id)
    pic = await generate_preview_pic(osu)
    await generate_preview.finish(MessageSegment.reply(event.message_id) + MessageSegment.image(pic))


generate_preview = on_command('完整预览', priority=11, block=True)


@generate_preview.handle()
async def _(event: Union[MessageEvent, GuildMessageEvent], msg: Message = CommandArg()):
    osu_id = msg.extract_plain_text().strip()
    if not osu_id or not osu_id.isdigit():
        await osudl.finish(MessageSegment.reply(event.message_id) + '请输入正确的地图mapID')
    data = await osu_api('map', map_id=int(osu_id))
    if not data:
        await generate_preview.finish(MessageSegment.reply(event.message_id) + '未查询到该地图')
    if isinstance(data, str):
        await generate_preview.finish(MessageSegment.reply(event.message_id) + data)
    osu = await download_tmp_osu(osu_id)
    pic = await generate_preview_pic(osu, full=True)
    await generate_preview.finish(MessageSegment.reply(event.message_id) + MessageSegment.image(pic))

update_pic = on_command('更新背景', aliases={'更改背景'}, priority=11, block=True)


@update_pic.handle(parameterless=[split_msg()])
async def _(bot: Bot, state: T_State, event: GroupMessageEvent, pic_ls: list = ImageURLs('请在指令后附上图片')):
    global counter
    if 'error' in state:
        await update_pic.finish(MessageSegment.reply(event.message_id) + state['error'])
    user = state['user']
    pic_url = pic_ls[0]
    review_pic_ls.append(ReviewData(event.message_id, pic_url, event.group_id, str(user), counter))
    msg = f'收到id为{counter}来自群{event.group_id}的更新背景申请' + MessageSegment.image(pic_url)
    for superuser in get_driver().config.superusers:
        await bot.send_private_msg(user_id=int(superuser), message=msg)
    counter += 1
    await update_pic.finish(MessageSegment.reply(event.message_id) + '已收到图片，请等待审核捏~')


update = on_command('update', aliases={'更新'}, priority=11, block=True)


@update.handle(parameterless=[split_msg()])
async def _(state: T_State, event: Union[MessageEvent, GuildMessageEvent]):
    if 'error' in state:
        await update.finish(MessageSegment.reply(event.message_id) + state['error'])
    user = state['user']
    path = user_cache_path / str(user) / 'icon.png'
    gif_path = user_cache_path / str(user) / 'icon.gif'
    if path.exists():
        path.unlink()
    if gif_path.exists():
        gif_path.unlink()
    await update.finish(MessageSegment.reply(event.message_id) + '个人信息更新成功')

accept = on_command('同意全部', priority=11, block=True, permission=SUPERUSER)


@accept.handle()
async def _():
    for i in review_pic_ls:
        await save_info_pic(i.user, i.pic_url)
    await accept.finish('全部审核通过')


reject = on_command('拒绝', aliases={'否决'}, priority=11, block=True, permission=SUPERUSER)


@reject.handle()
async def _(bot: Bot, msg: Message = CommandArg()):
    arg = msg.extract_plain_text().strip()
    for num, i in enumerate(review_pic_ls):
        if i.id == int(arg):
            msg = MessageSegment.reply(i.msg_id) + '你的提交的图片被拒绝，请重新上传'
            await bot.send_group_msg(group_id=i.group, message=msg)
            del review_pic_ls[num]
            break
    await reject.finish(f'拒绝id{arg}成功')


recommend = on_command('recommend', aliases={'推荐', '推荐铺面', '推荐谱面'}, priority=11, block=True)


@recommend.handle(parameterless=[split_msg()])
async def _(event: Union[MessageEvent, GuildMessageEvent], state: T_State):
    if 'error' in state:
        await recommend.finish(MessageSegment.reply(event.message_id) + state['error'])
    user = state['user']
    mode = state['mode']
    mods = state['mods']
    if mods == ['4K']:
        key_count = '4'
    elif mods == ['7K']:
        key_count = '7'
    else:
        key_count = '4,7'
    if mode == '1' or mode == '2':
        await recommend.finish('很抱歉，该模式暂不支持推荐')
    if not recommend_cache.get(user):
        recommend_cache[user] = set()
        await update_recommend(user)
    recommend_data = await get_recommend(user, mode, key_count)
    shuffle(recommend_data.data.list)
    if not recommend_data.data or not recommend_data.data.list:
        await recommend.finish('没有可以推荐的图哦，自己多打打喜欢玩的图吧')
    for i in recommend_data.data.list:
        if i.id not in recommend_cache[user]:
            recommend_cache[user].add(i.id)
            recommend_map = i
            break
    else:
        await recommend.finish('今天已经没有可以推荐的图啦，明天再来吧')
        return
    bid = int(re.findall('https://osu.ppy.sh/beatmaps/(.*)', recommend_map.mapLink)[0])
    map_info = await get_sayo_map_info(bid, 1)
    sid = map_info.data.sid
    for i in map_info.data.bid_data:
        if i.bid == bid:
            bg = i.bg
            break
    else:
        bg = ''
        logger.debug(f'如果看到这句话请联系作者 有问题的是{bid}, {sid}')
    s = f'推荐的铺面是{recommend_map.mapName} ⭐{round(recommend_map.difficulty, 2)}\n{"".join(recommend_map.mod)}\n' \
        f'预计pp为{round(recommend_map.predictPP, 2)}\n提升概率为{round(recommend_map.passPercent*100, 2)}%\n' \
        f'{recommend_map.mapLink}\nhttps://kitsu.moe/api/d/{sid}\nhttps://txy1.sayobot.cn/beatmaps/download/novideo/{sid}'
    await recommend.finish(MessageSegment.reply(event.message_id) +
                           MessageSegment.image(f'https://dl.sayobot.cn/beatmaps/files/{sid}/{bg}') + s)

osu_help = on_command('osuhelp', priority=11, block=True)
with open(Path(__file__).parent / 'osufile' / 'help.png', 'rb') as f:
    img1 = f.read()
with open(Path(__file__).parent / 'osufile' / 'detail.png', 'rb') as f:
    img2 = f.read()


@osu_help.handle()
async def _help(event: Union[MessageEvent, GuildMessageEvent], msg: Message = CommandArg()):
    arg = msg.extract_plain_text().strip()
    if not arg:
        await osu_help.finish(MessageSegment.reply(event.message_id) + MessageSegment.image(img1))
    if arg == 'detail':
        await osu_help.finish(MessageSegment.reply(event.message_id) + MessageSegment.image(img2))
    else:
        await osu_help.finish(MessageSegment.reply(event.message_id) + '呜呜，detail都打不对吗(ノ｀Д)ノ')

url_match = on_regex("https://osu.ppy.sh/beatmapsets/(.*)#")


@url_match.handle()
async def _url(event: Union[MessageEvent, GuildMessageEvent], bid: tuple = RegexGroup()):
    url_1 = "https://osu.direct/api/d/"
    url_2 = "https://txy1.sayobot.cn/beatmaps/download/novideo/"
    url_total = f"kitsu镜像站：{url_1}{bid[0]}\n小夜镜像站：{url_2}{bid[0]}"
    await url_match.finish(MessageSegment.reply(event.message_id) + url_total)


@scheduler.scheduled_job('cron', hour='0', misfire_grace_time=60)
async def update_info():
    result = await UserData.all()
    if not result:
        return
    for user in result:
        await update_user_info(user.osu_id)
    logger.info(f'已更新{len(result)}位玩家数据')


manual_update = on_command('更新数据', priority=11, block=True)


@manual_update.handle()
async def _():
    await update_info()


@scheduler.scheduled_job('cron', hour='4', day_of_week='0,4', misfire_grace_time=60)
async def delete_cached_map():
    map_path = Path('data/osu/map')
    shutil.rmtree(map_path)
    map_path.mkdir(parents=True, exist_ok=True)
    user_path = Path('data/osu/user')
    for file_path in user_path.glob('**/*'):
        if file_path.is_file() and file_path.name in ('icon.png', 'icon.gif'):
            file_path.unlink()



import random
from io import BytesIO
from httpx import AsyncClient, Response
from typing import Union, Optional
from nonebot.log import logger
from nonebot import get_driver
from expiringdict import ExpiringDict
from .config import Config
from .network import auto_retry
from .schema import SayoBeatmap, RecommendData

api = 'https://osu.ppy.sh/api/v2'
sayoapi = 'https://api.sayobot.cn'
pp_calc_api = 'https://api.yuzuai.xyz/osu/ppcalc'
cache = ExpiringDict(max_len=1, max_age_seconds=86400)
plugin_config = Config.parse_obj(get_driver().config.dict())
proxy = plugin_config.osu_proxy

if plugin_config.osu_key and plugin_config.osu_client:
    key = plugin_config.osu_key
    client_id = plugin_config.osu_client
else:
    raise Exception("请设置osu_key和osu_client")

if not plugin_config.info_bg:
    bg_url = ['https://api.ghser.com/random/pe.php']
else:
    bg_url = plugin_config.info_bg


@auto_retry
async def safe_async_get(url, headers: Optional[dict] = None, params: Optional[dict] = None) -> Response:
    async with AsyncClient(timeout=100, proxies=proxy, follow_redirects=True) as client:
        client: AsyncClient
        req = await client.get(url, headers=headers, params=params)
    return req


@auto_retry
async def safe_async_post(url, headers=None, data=None) -> Response:
    async with AsyncClient(timeout=100, proxies=proxy) as client:
        client: AsyncClient
        req = await client.post(url, headers=headers, data=data)
    return req


async def renew_token():
    url = "https://osu.ppy.sh/oauth/token"
    async with AsyncClient(timeout=100, proxies=proxy) as client:
        client: AsyncClient
        req = await client.post(url, json={'client_id': f'{client_id}', 'client_secret': f'{key}',
                                           'grant_type': 'client_credentials', 'scope': 'public'})
    if req.status_code == 200:
        osu_token = req.json()
        cache.update({
            'token': osu_token['access_token']
        })
    else:
        logger.error(f'更新OSU token出错 错误{req.status_code}')


async def osu_api(project: str, uid: int = 0, mode: str = 'osu', map_id: int = 0, is_name: bool = False) -> Union[str, dict]:
    if is_name:
        info = await get_user_info(f'{api}/users/{uid}?key=username')
        if isinstance(info, str):
            return info
        else:
            uid = info['id']
    if project == 'info' or project == 'bind' or project == 'update':
        url = f'{api}/users/{uid}/{mode}'
    elif project == 'recent':
        url = f'{api}/users/{uid}/scores/{project}?mode={mode}&include_fails=1'
    elif project == 'pr':
        url = f'{api}/users/{uid}/scores/recent?mode={mode}'
    elif project == 'score':
        url = f'{api}/beatmaps/{map_id}/scores/users/{uid}/all?mode={mode}'
    elif project == 'bp':
        url = f'{api}/users/{uid}/scores/best?mode={mode}&limit=100'
    elif project == 'map':
        url = f'{api}/beatmaps/{map_id}'
    else:
        raise 'Project Error'
    return await api_info(project, url)


async def sayo_api(setid: int) -> dict:
    url = f'{sayoapi}/v2/beatmapinfo?0={setid}'
    return await api_info('mapinfo', url)


async def get_user_info(url: str) -> Union[dict, str]:
    token = cache.get('token')
    if not token:
        await renew_token()
        token = cache.get('token')
    header = {'Authorization': f'Bearer {token}'}
    req = await safe_async_get(url, headers=header)
    if not req:
        return 'api请求失败，请稍后再试'
    if req.status_code == 401:
        await token.update_token()
        return await get_user_info(url)
    elif req.status_code == 404:
        return '未找到该玩家，请确认玩家ID是否正确，有无多余或缺少的空格'
    elif req.status_code == 200:
        return req.json()
    else:
        return 'API请求失败，请联系管理员'


async def api_info(project: str, url: str) -> Union[dict, str]:
    if project == 'mapinfo' or project == 'PPCalc':
        headers = {
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; WOW64) Chrome/78.0.3904.108'}
    else:
        token = cache.get('token')
        if not token:
            await renew_token()
            token = cache.get('token')
        headers = {'Authorization': f'Bearer {token}'}
    req = await safe_async_get(url, headers=headers)
    if not req:
        return 'api请求失败，请稍后再试'
    if req.status_code >= 400:
        if project == 'info' or project == 'bind':
            return '未找到该玩家，请确认玩家ID'
        elif project == 'recent':
            return '未找到该玩家，请确认玩家ID'
        elif project == 'score':
            return '未找到该地图成绩，请检查是否搞混了mapID与setID或模式'
        elif project == 'bp':
            return '未找到该玩家BP'
        elif project == 'map':
            return '未找到该地图，请检查是否搞混了mapID与setID'
        else:
            return 'API请求失败，请联系管理员或稍后再尝试'
    return req.json()


async def get_beatmap_attribute(map_id, mode):
    url = f'{api}/beatmaps/{map_id}/attributes'
    token = cache.get('token')
    if not token:
        await renew_token()
        token = cache.get('token')
    headers = {'Authorization': f'Bearer {token}'}
    req = await safe_async_post(url, headers, {'ruleset': mode})
    if req.status_code == 404:
        raise '内部错误'
    return req.json()


async def get_random_bg() -> Optional[bytes]:
    res = await safe_async_get(random.choice(bg_url))
    if res.status_code != 200:
        return
    return res.content


async def get_sayo_map_info(sid, t=0) -> SayoBeatmap:
    res = await safe_async_get(f'https://api.sayobot.cn/v2/beatmapinfo?K={sid}&T={t}')
    return SayoBeatmap(**res.json())


async def get_map_bg(sid, bg_name) -> BytesIO:
    res = await safe_async_get(f'https://dl.sayobot.cn/beatmaps/files/{sid}/{bg_name}')
    return BytesIO(res.content)


async def get_seasonal_bg() -> Optional[dict]:
    url = f'{api}/seasonal-backgrounds'
    token = cache.get('token')
    if not token:
        await renew_token()
        token = cache.get('token')
    headers = {'Authorization': f'Bearer {token}'}
    req = await safe_async_get(url, headers=headers)
    if req.status_code != 200:
        return
    return req.json()


async def get_recommend(uid, mode, key_count):
    headers = {'uid': str(uid)}
    params = {'newRecordPercent': '0.2,1',
              'passPercent': '0.2,1',
              'difficulty': '0,15',
              'keyCount': key_count,
              'gameMode': mode,
              'rule': '4',
              'current': '1',
              'pageSize': '20',
              'from': 'nonebot_plugin_osubot',
              'hidePlayed': '1'}
    res = await safe_async_get('https://alphaosu.keytoix.vip/api/v1/self/maps/recommend', headers=headers,
                               params=params)
    return RecommendData(**res.json())


async def update_recommend(uid):
    headers = {'uid': str(uid)}
    await safe_async_post('https://alphaosu.keytoix.vip/api/v1/self/users/synchronize', headers=headers)



from .score import Score, BeatmapUserScore
from .user import User, Badge
from .beatmap import Beatmap, Beatmapset, BeatmapDifficultyAttributes, SeasonalBackgrounds
from .sayo_beatmap import SayoBeatmap
from .alphaosu import RecommendData



import asyncio
import datetime
from io import BytesIO

from nonebot import on_command
from nonebot.adapters.onebot.v11 import GroupMessageEvent, MessageSegment, Bot
from nonebot.params import T_State
from PIL import Image, ImageDraw
from .utils import split_msg
from ..database.models import UserData, InfoData
from ..draw.static import Torus_Regular_25
from ..draw.utils import draw_fillet
from ..file import get_projectimg
from ..utils import NGM


group_pp_rank = on_command('群内排名', priority=11, block=True)


@group_pp_rank.handle(parameterless=[split_msg()])
async def _(event: GroupMessageEvent, state: T_State, bot: Bot):
    if 'error' in state:
        await group_pp_rank.finish(MessageSegment.reply(event.message_id) + state['error'])
    mode = state['mode']
    group_id = event.group_id
    group_member = await bot.get_group_member_list(group_id=group_id)
    user_id_ls = [i['user_id'] for i in group_member]
    binded_id = await UserData.filter(user_id__in=user_id_ls).values_list('osu_id', flat=True)
    info_ls = await InfoData.filter(osu_id__in=binded_id).filter(osu_mode=mode).filter(date=datetime.date.today()).order_by('-pp').all()
    icon_ls = [f'https://a.ppy.sh/{info.osu_id}' for info in info_ls]
    tasks = [get_projectimg(i) for i in icon_ls]
    icon_ls = await asyncio.gather(*tasks)
    draw_len = len([i for i in info_ls if i.pp >= 100])
    img = Image.new('RGBA', (1200, 85 + 82 * draw_len), (35, 42, 34, 255))
    draw = ImageDraw.Draw(img)
    draw.text((40, 10), f'{NGM[mode]}模式群内排名', font=Torus_Regular_25, fill=(255, 255, 255, 255))
    draw.text((880, 10), 'pp', font=Torus_Regular_25, fill=(255, 255, 255, 255))
    draw.text((960, 10), '全球排名', font=Torus_Regular_25, fill=(255, 255, 255, 255))
    draw.text((1070, 10), '国内排名', font=Torus_Regular_25, fill=(255, 255, 255, 255))
    for index, (info, icon) in enumerate(zip(info_ls, icon_ls)):
        if info.pp < 100:
            continue
        draw.rounded_rectangle((20, 55 + 82 * index, 1180, 45 + 82 * (index + 1)), radius=10, fill=(58, 70, 57, 255))
        icon_img = Image.open(icon).convert('RGBA').resize((63, 63))
        icon_img = draw_fillet(icon_img, 10)
        img.alpha_composite(icon_img, (100, 60 + 82 * index))
        for user in group_member:
            if await UserData.filter(user_id=user['user_id'], osu_id=info.osu_id).first():
                name = user['card'] or user.get('nickname', '')
                break
        else:
            raise Exception('这不可能发生的')
        draw.text((43, 70 + 82 * index), f'#{index + 1}', font=Torus_Regular_25, fill=(255, 255, 255, 255))
        draw.text((180, 70 + 82 * index), name, font=Torus_Regular_25, fill=(166, 199, 163, 255))
        draw.text((850, 70 + 82 * index), f'{int(info.pp)}pp', font=Torus_Regular_25, fill=(255, 255, 255, 255))
        draw.text((980, 70 + 82 * index), f'{info.g_rank}', font=Torus_Regular_25, fill=(255, 255, 255, 255))
        draw.text((1090, 70 + 82 * index), f'{info.c_rank}', font=Torus_Regular_25, fill=(255, 255, 255, 255))

    byt = BytesIO()
    img.save(byt, "png")
    await group_pp_rank.finish(MessageSegment.image(byt))



import asyncio
import io
from datetime import datetime, timedelta, date
from time import strptime, mktime
from typing import List, Union, Optional

from PIL import ImageDraw, UnidentifiedImageError
from nonebot.adapters.onebot.v11 import MessageSegment

from ..schema import Score
from ..api import osu_api
from ..utils import GMN
from ..mods import get_mods_list
from ..file import get_projectimg

from .utils import draw_fillet, draw_fillet2
from .static import *


async def draw_bp(project: str, uid: int, mode: str, mods: Optional[List], low_bound: int = 0, high_bound: int = 0,
                  day: int = 0, is_name: bool = False) -> Union[str, MessageSegment]:
    bp_info = await osu_api('bp', uid, mode, is_name=is_name)
    if isinstance(bp_info, str):
        return bp_info
    score_ls = [Score(**i) for i in bp_info]
    if not bp_info:
        return f'未查询到在 {GMN[mode]} 的游玩记录'
    user = bp_info[0]['user']['username']
    if project == 'bp':
        if mods:
            mods_ls = get_mods_list(score_ls, mods)
            if low_bound > len(mods_ls):
                return f'未找到开启 {"|".join(mods)} Mods的成绩'
            if high_bound > len(mods_ls):
                mods_ls = mods_ls[low_bound - 1:]
            else:
                mods_ls = mods_ls[low_bound - 1: high_bound]
            score_ls_filtered = [score_ls[i] for i in mods_ls]
        else:
            score_ls_filtered = score_ls[low_bound - 1: high_bound]
    else:
        ls = []
        for i, score in enumerate(score_ls):
            today = date.today() - timedelta(days=day)
            today_stamp = mktime(strptime(str(today), '%Y-%m-%d'))
            playtime = datetime.strptime(score.created_at.replace('Z', ''), '%Y-%m-%dT%H:%M:%S') + timedelta(hours=8)
            play_stamp = mktime(strptime(str(playtime), '%Y-%m-%d %H:%M:%S'))
            if play_stamp > today_stamp:
                ls.append(i)
        score_ls_filtered = [score_ls[i] for i in ls]
        if not score_ls_filtered:
            return f'近{day + 1}日内在 {GMN[mode]} 没有新增的BP成绩'
    msg = await draw_pfm(project, user, score_ls, score_ls_filtered, mode, low_bound, high_bound, day)
    return msg


async def draw_pfm(project: str, user: str, score_ls: List[Score], score_ls_filtered: List[Score],
                   mode: str, low_bound: int = 0, high_bound: int = 0,
                   day: int = 0) -> Union[str, MessageSegment]:
    task0 = [get_projectimg(f'https://assets.ppy.sh/beatmaps/{i.beatmapset.id}/covers/list.jpg')
             for i in score_ls_filtered]
    task1 = [get_projectimg(f'https://assets.ppy.sh/beatmaps/{i.beatmapset.id}/covers/cover.jpg')
             for i in score_ls_filtered]
    bg_ls = await asyncio.gather(*task0)
    large_banner_ls = await asyncio.gather(*task1)
    bplist_len = len(score_ls_filtered)
    im = Image.new('RGBA', (1420, 280 + 177 * ((bplist_len + 1) // 2 - 1)), (31, 41, 46, 255))
    im.alpha_composite(BgImg)
    draw = ImageDraw.Draw(im)
    f_div = Image.new('RGBA', (1450, 2), (255, 255, 255, 255)).convert('RGBA')
    im.alpha_composite(f_div, (0, 100))
    if project == 'bp':
        uinfo = f"玩家：{user} | {mode.capitalize()} 模式 | BP {low_bound} - {high_bound}"
    else:
        uinfo = f"玩家：{user} | {mode.capitalize()} 模式 | 近{day + 1}日新增 BP"
    draw.text((1388, 50), uinfo, font=Torus_SemiBold_25, anchor='rm')
    for num, bp in enumerate(score_ls_filtered):
        h_num = 177 * (num // 2)  # 每两个数据换行
        offset = 695 * (num % 2)  # 数据在右边时的偏移量

        # BP排名
        index = score_ls.index(bp)
        draw.text((40 + offset, 190 + h_num), str(index + 1), font=Torus_Regular_20, anchor='rm')

        # 获取谱面大banner
        try:
            banner = Image.open(large_banner_ls[num]).convert('RGBA').resize((650, 150))
            banner_with_fillet = draw_fillet2(banner, 15)
            im.alpha_composite(banner_with_fillet, (45 + offset, 114 + h_num))
        except UnidentifiedImageError:
            ...

        # 获取谱面banner
        try:
            bg = Image.open(bg_ls[num]).convert('RGBA').resize((150, 150))
            bg_imag = draw_fillet(bg, 15)
            im.alpha_composite(bg_imag, (45 + offset, 114 + h_num))
        except UnidentifiedImageError:
            ...

        # mods
        if bp.mods:
            for mods_num, s_mods in enumerate(bp.mods):
                mods_bg = osufile / 'mods' / f'{s_mods}.png'
                mods_img = Image.open(mods_bg).convert('RGBA')
                im.alpha_composite(mods_img, (210 + offset + 50 * mods_num, 192 + h_num))
            if (bp.rank == 'X' or bp.rank == 'S') and ('HD' in bp.mods or 'FL' in bp.mods):
                bp.rank += 'H'

        # 曲名&作曲
        metadata = f'{bp.beatmapset.title} | by {bp.beatmapset.artist}'
        if len(metadata) > 30:
            metadata = metadata[:27] + '...'
        draw.text((210 + offset, 135 + h_num), metadata, font=Torus_Regular_25, anchor='lm')

        # 地图版本&时间
        old_time = datetime.strptime(bp.created_at.replace('Z', ''), '%Y-%m-%dT%H:%M:%S')
        new_time = (old_time + timedelta(hours=8)).strftime('%Y-%m-%d %H:%M:%S')
        difficulty = bp.beatmap.version
        if len(difficulty) > 30:
            difficulty = difficulty[:27] + '...'
        difficulty = f'{bp.beatmap.difficulty_rating}★ | {difficulty}'
        draw.text((210 + offset, 168 + h_num), difficulty, font=Torus_Regular_20, anchor='lm', fill=(238, 171, 0, 255))

        # 达成时间
        draw.text((210 + offset, 245 + h_num), f'{new_time}', font=Torus_Regular_20, anchor='lm')

        # acc
        draw.text((658 + offset, 220 + h_num), f'{bp.accuracy * 100:.2f}%', font=Torus_SemiBold_20,
                  anchor='rm', fill=(238, 171, 0, 255))

        # rank
        rank_img = osufile / 'ranking' / f'ranking-{bp.rank}.png'
        rank_bg = Image.open(rank_img).convert('RGBA').resize((32, 16))
        im.alpha_composite(rank_bg, (660 + offset, 213 + h_num))

        # mapid
        draw.text((690 + offset, 245 + h_num), f'ID: {bp.beatmap.id}', font=Torus_Regular_20, anchor='rm')

        # pp
        draw.text((647 + offset, 184 + h_num), f'{bp.pp:.0f}', font=Torus_SemiBold_25,
                  anchor='rm', fill=(255, 102, 171, 255))
        draw.text((650 + offset, 180 + h_num), 'pp', font=Torus_SemiBold_25,
                  anchor='lm', fill=(209, 148, 176, 255))

        # 分割线
        div = Image.new('RGBA', (1400, 2), (46, 53, 56, 255)).convert('RGBA')
        im.alpha_composite(div, (25, 275 + h_num))

        await asyncio.sleep(0)

    image_bytesio = io.BytesIO()
    im.save(image_bytesio, 'PNG')
    # 转换为 JPEG 格式
    image_bytesio.seek(0)
    image = Image.open(image_bytesio).convert('RGB')
    image_bytesio = io.BytesIO()
    image.save(image_bytesio, 'JPEG', quality=85)
    image.close()
    msg = MessageSegment.image(image_bytesio.getvalue())
    image_bytesio.close()

    return msg



from datetime import datetime, timedelta
from typing import Union

from PIL import ImageFilter, ImageEnhance, ImageDraw
from nonebot.adapters.onebot.v11 import MessageSegment

from ..api import osu_api, sayo_api
from ..schema import SayoBeatmap
from ..file import get_projectimg
from .static import *
from .utils import crop_bg, calc_songlen, stars_diff, image2bytesio


async def draw_bmap_info(mapid, op: bool = False) -> Union[str, MessageSegment]:
    if op:
        info = await osu_api('map', map_id=mapid)
        if not info:
            return '未查询到地图'
        elif isinstance(info, str):
            return info
        mapid = info['beatmapset_id']
    info = await sayo_api(mapid)
    if isinstance(info, str):
        return info
    sayo_info = SayoBeatmap(**info)
    if sayo_info.status == -1:
        return '在sayobot未查询到该地图'
    data = sayo_info.data

    coverurl = f'https://assets.ppy.sh/beatmaps/{mapid}/covers/cover@2x.jpg'
    cover = await get_projectimg(coverurl)
    # 新建
    if len(data.bid_data) > 20:
        im_h = 400 + 102 * 20
    else:
        im_h = 400 + 102 * (len(data.bid_data) - 1)
    im = Image.new('RGBA', (1200, im_h), (31, 41, 46, 255))
    draw = ImageDraw.Draw(im)
    # 背景
    cover_crop = await crop_bg('MP', cover)
    cover_gb = cover_crop.filter(ImageFilter.GaussianBlur(1))
    cover_img = ImageEnhance.Brightness(cover_gb).enhance(2 / 4.0)
    im.alpha_composite(cover_img, (0, 0))
    # 曲名
    draw.text((25, 40), data.title, font=Torus_SemiBold_40, anchor='lt')
    # 曲师
    draw.text((25, 75), f'by {data.artist}', font=Torus_SemiBold_20, anchor='lt')
    # mapper
    draw.text((25, 110), f'谱面作者: {data.creator}', font=Torus_SemiBold_20, anchor='lt')
    # rank时间
    if data.approved_date == -1:
        approved_date = '谱面状态可能非ranked'
    else:
        datearray = datetime.utcfromtimestamp(data.approved_date)
        approved_date = (datearray + timedelta(hours=8)).strftime('%Y-%m-%d %H:%M:%S')
    draw.text((25, 145), f'上架时间: {approved_date}', font=Torus_SemiBold_20, anchor='lt')
    # 来源
    draw.text((25, 180), f'Source: {data.source}', font=Torus_SemiBold_20, anchor='lt')
    # bpm
    draw.text((1150, 110), f'BPM: {data.bpm}', font=Torus_SemiBold_20, anchor='rt')
    # 曲长
    music_len = calc_songlen(data.bid_data[0].length)
    draw.text((1150, 145), f'lenght: {music_len}', font=Torus_SemiBold_20, anchor='rt')
    # Setid
    draw.text((1150, 20), f'Setid: {mapid}', font=Torus_SemiBold_20, anchor='rt')
    gmap = sorted(data.bid_data, key=lambda k: k.star, reverse=False)
    for num, cmap in enumerate(gmap):
        if num < 20:
            h_num = 102 * num
            # 难度
            mode_bg = stars_diff(cmap.mode, cmap.star)
            mode_img = mode_bg.resize((20, 20))
            im.alpha_composite(mode_img, (20, 320 + h_num))
            # 星星
            stars_bg = stars_diff('stars', cmap.star)
            stars_img = stars_bg.resize((20, 20))
            im.alpha_composite(stars_img, (50, 320 + h_num))
            # diff
            im.alpha_composite(BarImg, (10, 365 + h_num))
            gc = ['CS', 'HP', 'OD', 'AR']
            for index, i in enumerate((cmap.CS, cmap.HP, cmap.OD, cmap.AR)):
                diff_len = int(200 * i / 10) if i <= 10 else 200
                diff_bg = Image.new('RGBA', (diff_len, 12), (255, 255, 255, 255))
                im.alpha_composite(diff_bg, (50 + 300 * index, 365 + h_num))
                draw.text((20 + 300 * index, 369 + h_num), gc[index], font=Torus_SemiBold_20, anchor='lm')
                draw.text((265 + 300 * index, 369 + h_num), f'{i}', font=Torus_SemiBold_20, anchor='lm', fill=(255, 204, 34, 255))
                if index != 3:
                    draw.text((300 + 300 * index, 369 + h_num), '|', font=Torus_SemiBold_20, anchor='lm')
            # 难度
            draw.text((80, 328 + h_num),  f'{cmap.star:.1f}', font=Torus_SemiBold_20, anchor='lm')
            # version
            draw.text((125, 328 + h_num), f' |  {cmap.version}', font=Torus_SemiBold_20, anchor='lm')
            # mapid
            draw.text((1150, 328 + h_num), f'Mapid: {cmap.bid}', font=Torus_SemiBold_20, anchor='rm')
            # maxcb
            draw.text((700, 328 + h_num), f'Max Combo: {cmap.maxcombo}', font=Torus_SemiBold_20, anchor='lm')
            # 分割线
            div = Image.new('RGBA', (1150, 2), (46, 53, 56, 255)).convert('RGBA')
            im.alpha_composite(div, (25, 400 + h_num))
        else:
            plusnum = f'+ {num - 19}'
            draw.text((600, 350 + 102 * 20), plusnum, font=Torus_SemiBold_50, anchor='mm')
    base = image2bytesio(im)
    im.close()
    msg = MessageSegment.image(base)
    return msg



from datetime import date
from nonebot.log import logger
from ..utils import GM
from ..database.models import UserData, InfoData
from ..schema import User
from ..api import osu_api


async def bind_user_info(project: str, uid, qid, is_name) -> str:
    info = await osu_api(project, uid, GM[0], is_name=is_name)
    if not info:
        return f'未查询到玩家“{uid}”，请检查是否有多于或缺少的空格'
    elif isinstance(info, str):
        return info
    uid = info['id']
    name = info['username']
    await UserData.create(user_id=qid, osu_id=uid, osu_name=name, osu_mode=0)
    await update_user_info(uid)
    msg = f'用户 {name} 已成功绑定QQ {qid}\n默认模式为std，若更改模式至mania请输入/更新模式 3'
    return msg


async def update_user_info(uid: int):
    for mode in range(4):
        userinfo_dic = await osu_api('update', uid, GM[mode])
        if isinstance(userinfo_dic, str):
            logger.error(f'获取uid{uid} 更新信息出错\n {userinfo_dic}')
            continue
        userinfo = User(**userinfo_dic)
        if await InfoData.filter(osu_id=uid, osu_mode=mode, date=date.today()).first():
            continue
        elif userinfo.statistics.play_count:
            await InfoData.create(osu_id=uid,
                                  c_rank=userinfo.statistics.country_rank,
                                  g_rank=userinfo.statistics.global_rank,
                                  pp=userinfo.statistics.pp,
                                  acc=round(userinfo.statistics.hit_accuracy, 2),
                                  pc=userinfo.statistics.play_count,
                                  count=userinfo.statistics.total_hits,
                                  osu_mode=mode,
                                  date=date.today())
        else:
            await InfoData.create(osu_id=uid,
                                  c_rank=0,
                                  g_rank=0,
                                  pp=0,
                                  acc=0,
                                  pc=0,
                                  count=0,
                                  osu_mode=mode,
                                  date=date.today())
    logger.info(f'玩家:[{uid}] 个人信息更新完毕')



from dataclasses import dataclass
from io import BytesIO
from typing import Optional

from reamber.algorithms.generate import full_ln
from reamber.osu.OsuHit import OsuHit
from reamber.osu.OsuMap import OsuMap
from reamber.algorithms.playField import PlayField
from reamber.algorithms.playField.parts import *
from reamber.algorithms.pattern.combos.PtnCombo import PtnCombo
from reamber.algorithms.pattern.Pattern import Pattern
from pathlib import Path
from zipfile import ZipFile
from ..file import download_map
from ..schema import SayoBeatmap
import os
import shutil
import asyncio
import numpy as np

osu_path = Path() / "data" / "osu"
if not osu_path.exists():
    osu_path.mkdir(parents=True, exist_ok=True)


@dataclass
class Options:
    rate: Optional[float]
    end_rate: Optional[float]
    od: Optional[float]
    set: Optional[int]
    map: Optional[int] = None
    sayo_info: Optional[SayoBeatmap] = None
    nsv: bool = False
    nln: bool = False
    fln: bool = False
    step: float = 0.05
    gap: float = 150
    thres: float = 100


async def generate_preview_pic(file: Path, full=False) -> BytesIO:
    m = OsuMap.read_file(str(file.absolute()))
    keys = m.stack().column.max() + 1
    ptn = Pattern.from_note_lists([m.hits, m.holds], include_tails=False)
    grp = ptn.group()
    pf = (
            PlayField(m, duration_per_px=5, padding=60)
            + PFDrawColumnLines()
            + PFDrawBeatLines()
            + PFDrawBpm(x_offset=30)
            + PFDrawSv()
            + PFDrawNotes()
            + PFDrawOffsets(interval=2000, decimal_places=0)
    )
    if full:
        pf += PFDrawLines.from_combo(
                **PFDrawLines.Colors.RED,
                keys=keys,
                combo=np.concatenate(PtnCombo(grp).template_chord_stream(
                    primary=3, secondary=2,
                    keys=keys, and_lower=True
                ), axis=0)
            )
        pf += PFDrawLines.from_combo(
                **PFDrawLines.Colors.PURPLE,
                keys=keys,
                combo=np.concatenate(PtnCombo(grp).template_jacks(
                    minimum_length=2, keys=keys), axis=0)
            )
    byt = BytesIO()
    pf.export_fold(max_height=3000).save(byt, "png")
    return byt


async def convert_mania_map(options: Options) -> Optional[Path]:
    path = osu_path / f"{options.set}"
    osz_file = await download_map(options.set)
    if not osz_file:
        return
    with ZipFile(osz_file.absolute()) as my_zip:
        my_zip.extractall(path)
    os.remove(osz_file)
    if options.sayo_info:
        for i in options.sayo_info.data.bid_data:
            if i.bid == options.map:
                audio_file_name = i.audio
                audio_name = audio_file_name[:-4]
                audio_type = audio_file_name[-4:]
                break
        else:
            raise Exception('小夜api有问题啊')
    if options.rate:
        if options.rate > 10:
            options.rate = 10
        end = options.end_rate if options.end_rate else options.rate + 0.01
        if end > 10:
            end = 10.1
        if options.step and abs(options.step) < 0.05:
            options.step = 0.05 if options.step > 0 else -0.05
        if not options.step:
            options.step = 0.05
        tasks = []
        for rate in np.arange(options.rate, end, options.step):
            new_audio_path = path / (audio_name + f'x{rate:.2f}' + audio_type)
            tasks.append(asyncio.create_subprocess_shell(
                f'ffmpeg -i "{(path / audio_file_name).absolute()}" -filter:a "atempo={rate}" -b:a 128k -vn -y '
                f'"{new_audio_path.absolute()}" -loglevel quiet'
            ))
        await asyncio.gather(*tasks)
    osu_ls = list()
    for file in path.rglob('*.osu'):
        osu = OsuMap.read_file(str(file.absolute()))
        if options.rate:
            if osu.audio_file_name != audio_file_name:
                continue
            for rate in np.arange(options.rate, end, options.step):
                rate = round(rate, 2)
                osu_new = osu.rate(rate)
                osu_new.version += f' x{rate}'
                osu_new.audio_file_name = audio_name + f'x{rate:.2f}' + audio_type
                osu_ls.append([file.stem + f'x{rate}', osu_new])
        else:
            osu_new = osu.rate(1)
            osu_ls.append([file.stem, osu_new])
    for i in osu_ls:
        if options.fln:
            i[1] = full_ln(i[1], gap=options.gap, ln_as_hit_thres=options.thres)
            i[1].version += ' (FULL LN)'
            i[0] += ' (FULL LN)'
        if options.nsv:
            i[1].svs = i[1].svs[:0]
            i[1].bpms = i[1].bpms[:1]
            i[1].version += " NSV"
            i[0] += " NSV"
        if options.nln:
            for ln in i[1].holds:
                i[1].hits = i[1].hits.append(OsuHit(ln.offset, int(ln.column)))
            i[1].holds.df = i[1].holds.df[:0]
            i[1].version += " NLN"
            i[0] += " NLN"
        if options.od is not None:
            i[1].overall_difficulty = options.od
            i[1].version += f" OD {options.od}"
            i[0] += f" od{options.od}"
    for filename, osu in osu_ls:
        osu.write_file(path / f"{filename}.osu")

    with ZipFile(path.parent / osz_file.name, 'w') as my_zip:
        for file in path.rglob('*'):
            my_zip.write(file, os.path.relpath(file, path))
    shutil.rmtree(path)
    return path.parent / osz_file.name



import asyncio
from datetime import datetime, timedelta
from io import BytesIO
from typing import Optional, List, Union
from PIL import ImageFilter, ImageEnhance, ImageDraw, ImageSequence
from nonebot.adapters.onebot.v11 import MessageSegment

from ..api import osu_api, get_map_bg, get_beatmap_attribute, get_sayo_map_info
from ..schema import Score, Beatmap, User, BeatmapDifficultyAttributes
from ..mods import get_mods_list
from ..file import re_map, download_osu, user_cache_path, map_path
from ..utils import GMN, FGM
from ..pp import cal_pp, get_if_pp_ss_pp

from .static import *
from .utils import draw_acc, draw_fillet, crop_bg, calc_songlen, stars_diff, get_modeimage, open_user_icon


async def draw_score(project: str,
                     uid: int,
                     mode: str,
                     mods: Optional[List[str]],
                     best: int = 0,
                     mapid: int = 0,
                     is_name: bool = False) -> Union[str, MessageSegment]:
    task0 = asyncio.create_task(osu_api(project, uid, mode, mapid, is_name=is_name))
    task1 = asyncio.create_task(osu_api('info', uid, mode, is_name=is_name))
    score_json = await task0
    if not score_json:
        return f'未查询到在 {GMN[mode]} 的游玩记录'
    elif isinstance(score_json, str):
        return score_json
    if project in ('recent', 'pr'):
        if len(score_json) < best:
            return f'未查询到24小时内在 {GMN[mode]} 中第{best + 1}个游玩记录'
        score_info = Score(**score_json[best])
        grank = '--'
    elif project == 'bp':
        score_ls = [Score(**i) for i in score_json]
        mods_ls = get_mods_list(score_ls, mods)
        if len(mods_ls) < best:
            return f'未找到开启 {"|".join(mods)} Mods的第{best}个成绩'
        score_info = Score(**score_json[mods_ls[best - 1]])
        grank = '--'
    else:
        raise 'Project Error'
    # 从官网获取信息
    path = map_path / str(score_info.beatmap.beatmapset_id)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    osu = path / f"{score_info.beatmap.id}.osu"
    task = asyncio.create_task(get_beatmap_attribute(score_info.beatmap.id, mode))
    task2 = asyncio.create_task(osu_api('map', map_id=score_info.beatmap.id))
    if not osu.exists():
        await download_osu(score_info.beatmap.beatmapset_id, score_info.beatmap.id)
    info_json = await task1
    info = User(**info_json)
    user_path = user_cache_path / str(info.id)
    if not user_path.exists():
        user_path.mkdir(parents=True, exist_ok=True)
    map_json = await task2
    map_attribute_json = await task
    return await draw_score_pic(score_info, info, map_json, map_attribute_json,
                                score_info.beatmap.id, score_info.beatmap.beatmapset_id)


async def get_score_data(uid: int, mode: str, mods: Optional[List[str]], mapid: int = 0, is_name: bool = False):
    task = asyncio.create_task(get_beatmap_attribute(mapid, mode))
    task0 = asyncio.create_task(osu_api('score', uid, mode, mapid, is_name=is_name))
    task1 = asyncio.create_task(osu_api('info', uid, mode, is_name=is_name))
    task2 = asyncio.create_task(osu_api('map', map_id=mapid))
    task3 = asyncio.create_task(get_sayo_map_info(mapid, 1))
    score_json = await task0
    if not score_json:
        return f'未查询到在 {GMN[mode]} 的游玩记录'
    elif isinstance(score_json, str):
        return score_json
    if 'score' not in score_json:
        score_ls = [Score(**i) for i in score_json['scores']]
        if not score_ls:
            return f'未查询到在 {GMN[mode]} 的游玩记录'
        if mods:
            for score in score_ls:
                if mods == 'NM' and not score.mods:
                    score_info = score
                    break
                if score.mods == mods:
                    score_info = score
                    break
            else:
                for score in score_ls:
                    if set(mods).issubset(set(score.mods)):
                        score_info = score
                        break
                else:
                    return f'未找到开启 {"|".join(mods)} Mods的成绩'
        else:
            score_info = score_ls[0]
    else:
        score_info = Score(**score_json['score'])
    sayo_map_info = await task3
    path = map_path / str(sayo_map_info.data.sid)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    osu = path / f"{mapid}.osu"
    if not osu.exists():
        await download_osu(sayo_map_info.data.sid, mapid)
    info_json = await task1
    info = User(**info_json)
    user_path = user_cache_path / str(info.id)
    if not user_path.exists():
        user_path.mkdir(parents=True, exist_ok=True)
    map_json = await task2
    map_attribute_json = await task
    return await draw_score_pic(score_info, info, map_json, map_attribute_json, mapid, sayo_map_info.data.sid)


async def draw_score_pic(score_info, info, map_json, map_attribute_json, bid, sid):
    path = map_path / str(sid)
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    # pp
    osu = path / f"{bid}.osu"
    pp_info = cal_pp(score_info, str(osu.absolute()))
    if_pp, ss_pp = get_if_pp_ss_pp(score_info, str(osu.absolute()))
    # 新建图片
    im = Image.new('RGBA', (1500, 720))
    draw = ImageDraw.Draw(im)
    # 获取cover并裁剪，高斯，降低亮度
    cover = re_map(osu)
    cover_path = path / cover
    if not cover_path.exists():
        bg = await get_map_bg(sid, cover)
        with open(cover_path, 'wb') as f:
            f.write(bg.getvalue())
    cover_crop = await crop_bg('BG', cover_path)
    cover_gb = cover_crop.filter(ImageFilter.GaussianBlur(3))
    cover_img = ImageEnhance.Brightness(cover_gb).enhance(2 / 4.0)
    im.alpha_composite(cover_img, (0, 0))
    # 获取成绩背景做底图
    bg = get_modeimage(FGM[score_info.mode])
    recent_bg = Image.open(bg).convert('RGBA')
    im.alpha_composite(recent_bg)
    # 模式
    mode_bg = stars_diff(FGM[score_info.mode], pp_info.difficulty.stars)
    mode_img = mode_bg.resize((30, 30))
    im.alpha_composite(mode_img, (75, 74))
    # 难度星星
    stars_bg = stars_diff('stars', pp_info.difficulty.stars)
    stars_img = stars_bg.resize((23, 23))
    im.alpha_composite(stars_img, (134, 78))
    # mods
    if 'HD' in score_info.mods or 'FL' in score_info.mods:
        ranking = ['XH', 'SH', 'A', 'B', 'C', 'D', 'F']
    else:
        ranking = ['X', 'S', 'A', 'B', 'C', 'D', 'F']
    if score_info.mods:
        for mods_num, s_mods in enumerate(score_info.mods):
            mods_bg = osufile / 'mods' / f'{s_mods}.png'
            mods_img = Image.open(mods_bg).convert('RGBA')
            im.alpha_composite(mods_img, (500 + 50 * mods_num, 160))
    # 成绩S-F
    rank_ok = False
    for rank_num, i in enumerate(ranking):
        rank_img = osufile / 'ranking' / f'ranking-{i}.png'
        if rank_ok:
            rank_b = Image.open(rank_img).convert('RGBA').resize((48, 24))
            rank_new = Image.new('RGBA', rank_b.size, (0, 0, 0, 0))
            rank_bg = Image.blend(rank_new, rank_b, 0.5)
        elif i != score_info.rank:
            rank_b = Image.open(rank_img).convert('RGBA').resize((48, 24))
            rank_new = Image.new('RGBA', rank_b.size, (0, 0, 0, 0))
            rank_bg = Image.blend(rank_new, rank_b, 0.2)
        else:
            rank_bg = Image.open(rank_img).convert('RGBA').resize((48, 24))
            rank_ok = True
        im.alpha_composite(rank_bg, (75, 163 + 39 * rank_num))
    # 成绩+acc
    im = draw_acc(im, score_info.accuracy, score_info.mode)
    # 地区
    country = osufile / 'flags' / f'{info.country_code}.png'
    country_bg = Image.open(country).convert('RGBA').resize((66, 45))
    im.alpha_composite(country_bg, (250, 577))
    # supporter
    if info.is_supporter:
        im.alpha_composite(SupporterBg.resize((40, 40)), (250, 640))
    mapinfo = Beatmap(**map_json)
    map_attribute = BeatmapDifficultyAttributes(**map_attribute_json['attributes'])
    # cs, ar, od, hp, stardiff
    mapdiff = [mapinfo.cs, mapinfo.drain, mapinfo.accuracy, mapinfo.ar, pp_info.difficulty.stars]
    for num, i in enumerate(mapdiff):
        color = (255, 255, 255, 255)
        if num == 4:
            color = (255, 204, 34, 255)
        diff_len = int(250 * i / 10) if i <= 10 else 250
        diff_len = Image.new('RGBA', (diff_len, 8), color)
        im.alpha_composite(diff_len, (1190, 306 + 35 * num))
        draw.text((1470, 306 + 35 * num), f'{i:.1f}', font=Torus_SemiBold_20, anchor='mm')
    # 时长 - 滑条
    diff_info = calc_songlen(mapinfo.total_length), mapinfo.bpm, mapinfo.count_circles, mapinfo.count_sliders
    for num, i in enumerate(diff_info):
        draw.text((1070 + 120 * num, 245), f'{i}', font=Torus_Regular_20, anchor='lm', fill=(255, 204, 34, 255))
    # 状态
    draw.text((1400, 184), mapinfo.status.capitalize(), font=Torus_SemiBold_20, anchor='mm')
    # mapid
    draw.text((1425, 89), f'Mapid: {bid}', font=Torus_SemiBold_25, anchor='rm')
    # 曲名
    draw.text((75, 38), f'{mapinfo.beatmapset.title} | by {mapinfo.beatmapset.artist_unicode}', font=Torus_SemiBold_30,
              anchor='lm')
    # 星级
    draw.text((162, 89), f'{pp_info.difficulty.stars:.1f}', font=Torus_SemiBold_20, anchor='lm')
    # 谱面版本，mapper
    draw.text((225, 89), f'{mapinfo.version} | 谱师: {mapinfo.beatmapset.creator}', font=Torus_SemiBold_20, anchor='lm')
    # 评价
    draw.text((316, 307), score_info.rank, font=Venera_75, anchor='mm')
    # 分数
    draw.text((498, 251), f'{score_info.score:,}', font=Torus_Regular_75, anchor='lm')
    # 时间
    draw.text((498, 341), '达成时间:', font=Torus_SemiBold_20, anchor='lm')
    old_time = datetime.strptime(score_info.created_at.replace('Z', ''), '%Y-%m-%dT%H:%M:%S')
    new_time = (old_time + timedelta(hours=8)).strftime('%Y-%m-%d %H:%M:%S')
    draw.text((630, 341), new_time, font=Torus_SemiBold_20, anchor='lm')
    # 全球排名
    # draw.text((583, 410), grank, font=Torus_SemiBold_25, anchor='mm')
    # 左下玩家名
    draw.text((250, 530), info.username, font=Torus_SemiBold_30, anchor='lm')
    # 国内排名
    draw.text((325, 610), f'#{info.statistics.country_rank}', font=Torus_SemiBold_25, anchor='lm')
    if score_info.mode == 'osu':
        draw.text((720, 550), ss_pp, font=Torus_Regular_30, anchor='mm')
        draw.text((840, 550), if_pp, font=Torus_Regular_30, anchor='mm')
        draw.text((960, 550), f'{pp_info.pp:.0f}', font=Torus_Regular_30, anchor='mm')
        draw.text((720, 645), f'{pp_info.pp_aim:.0f}', font=Torus_Regular_30, anchor='mm')
        draw.text((840, 645), f'{pp_info.pp_speed:.0f}', font=Torus_Regular_30, anchor='mm')
        draw.text((960, 645), f'{pp_info.pp_acc:.0f}', font=Torus_Regular_30, anchor='mm')
        draw.text((1157, 550), f'{score_info.accuracy * 100:.2f}%', font=Torus_Regular_30, anchor='mm')
        draw.text((1385, 550), f'{score_info.max_combo:,}/{map_attribute.max_combo:,}', font=Torus_Regular_30,
                  anchor='mm')
        draw.text((1100, 645), f'{score_info.statistics.count_300}', font=Torus_Regular_30, anchor='mm')
        draw.text((1214, 645), f'{score_info.statistics.count_100}', font=Torus_Regular_30, anchor='mm')
        draw.text((1328, 645), f'{score_info.statistics.count_50}', font=Torus_Regular_30, anchor='mm')
        draw.text((1442, 645), f'{score_info.statistics.count_miss}', font=Torus_Regular_30, anchor='mm')
    elif score_info.mode == 'taiko':
        draw.text((1118, 550), f'{score_info.accuracy * 100:.2f}%', font=Torus_Regular_30, anchor='mm')
        draw.text((1270, 550), f'{score_info.max_combo:,}', font=Torus_Regular_30, anchor='mm')
        draw.text((1420, 550), f'{pp_info.pp:.0f}/{ss_pp}', font=Torus_Regular_30, anchor='mm')
        draw.text((1118, 645), f'{score_info.statistics.count_300}', font=Torus_Regular_30, anchor='mm')
        draw.text((1270, 645), f'{score_info.statistics.count_100}', font=Torus_Regular_30, anchor='mm')
        draw.text((1420, 645), f'{score_info.statistics.count_miss}', font=Torus_Regular_30, anchor='mm')
    elif score_info.mode == 'fruits':
        draw.text((1083, 550), f'{score_info.accuracy * 100:.2f}%', font=Torus_Regular_30, anchor='mm')
        draw.text((1247, 550), f'{score_info.max_combo}/{map_attribute.max_combo}', font=Torus_Regular_30, anchor='mm')
        draw.text((1411, 550), f'{pp_info.pp:.0f}/{ss_pp}', font=Torus_Regular_30, anchor='mm')
        draw.text((1062, 645), f'{score_info.statistics.count_300}', font=Torus_Regular_30, anchor='mm')
        draw.text((1185, 645), f'{score_info.statistics.count_100}', font=Torus_Regular_30, anchor='mm')
        draw.text((1309, 645), f'{score_info.statistics.count_katu}', font=Torus_Regular_30, anchor='mm')
        draw.text((1432, 645), f'{score_info.statistics.count_miss}', font=Torus_Regular_30, anchor='mm')
    else:
        draw.text((1002, 580), f'{score_info.statistics.count_geki / score_info.statistics.count_300 :.1f}:1'
        if score_info.statistics.count_300 != 0 else '∞:1', font=Torus_Regular_20, anchor='mm')
        draw.text((1002, 550), f'{score_info.accuracy * 100:.2f}%', font=Torus_Regular_30, anchor='mm')
        draw.text((1197, 550), f'{score_info.max_combo}', font=Torus_Regular_30, anchor='mm')
        draw.text((1395, 550), f'{pp_info.pp:.0f}/{ss_pp}', font=Torus_Regular_30, anchor='mm')
        draw.text((953, 645), f'{score_info.statistics.count_geki}', font=Torus_Regular_30, anchor='mm')
        draw.text((1051, 645), f'{score_info.statistics.count_300}', font=Torus_Regular_30, anchor='mm')
        draw.text((1150, 645), f'{score_info.statistics.count_katu}', font=Torus_Regular_30, anchor='mm')
        draw.text((1249, 645), f'{score_info.statistics.count_100}', font=Torus_Regular_30, anchor='mm')
        draw.text((1347, 645), f'{score_info.statistics.count_50}', font=Torus_Regular_30, anchor='mm')
        draw.text((1445, 645), f'{score_info.statistics.count_miss}', font=Torus_Regular_30, anchor='mm')
    user_icon = await open_user_icon(info)
    gif_frames = []
    if not getattr(user_icon, "is_animated", False):
        icon_bg = user_icon.convert('RGBA').resize((170, 170))
        icon_img = draw_fillet(icon_bg, 15)
        im.alpha_composite(icon_img, (60, 510))
        byt = BytesIO()
        im.save(byt, "png")
        msg = MessageSegment.image(byt)
        im.close()
        user_icon.close()
        return msg
    for gif_frame in ImageSequence.Iterator(user_icon):
        # 将 GIF 图片中的每一帧转换为 RGBA 模式
        gif_frame = gif_frame.convert('RGBA').resize((170, 170))
        gif_frame = draw_fillet(gif_frame, 15)
        # 创建一个新的 RGBA 图片，将 PNG 图片作为背景，将当前帧添加到背景上
        rgba_frame = Image.new('RGBA', im.size, (0, 0, 0, 0))
        rgba_frame.paste(im, (0, 0), im)
        rgba_frame.paste(gif_frame, (60, 510), gif_frame)
        # 将 RGBA 图片转换为 RGB 模式，并添加到 GIF 图片中
        gif_frames.append(rgba_frame)
    gif_bytes = BytesIO()
    # 保存 GIF 图片
    gif_frames[0].save(gif_bytes, format='gif', save_all=True, append_images=gif_frames[1:])
    # 输出
    gif_frames[0].close()
    user_icon.close()
    msg = MessageSegment.image(gif_bytes)
    return msg



from datetime import datetime, timedelta, date
from io import BytesIO
from typing import Union
from PIL import ImageDraw, ImageSequence
from nonebot.adapters.onebot.v11 import MessageSegment

from .static import *
from .utils import draw_fillet, info_calc, open_user_icon

from ..api import osu_api, get_random_bg
from ..schema import User
from ..file import make_badge_cache_file, user_cache_path, badge_cache_path
from ..database.models import InfoData
from ..utils import GMN, FGM


async def draw_info(uid: Union[int, str], mode: str, day: int, is_name) -> Union[str, MessageSegment]:
    info_json = await osu_api('info', uid, mode, is_name=is_name)
    if isinstance(info_json, str):
        return info_json
    info = User(**info_json)
    statistics = info.statistics
    if statistics.play_count == 0:
        return f'此玩家尚未游玩过{GMN[mode]}模式'
    # 对比
    user = await InfoData.filter(osu_id=info.id, osu_mode=FGM[mode]).order_by('-date').first()
    if user:
        today_date = date.today()
        query_date = today_date - timedelta(days=day)
        if user := await InfoData.filter(osu_id=info.id, osu_mode=FGM[mode], date__gte=query_date).order_by('date').first():
            n_crank, n_grank, n_pp, n_acc, n_pc, n_count = user.c_rank, user.g_rank, user.pp, user.acc, user.pc, user.count
        else:
            n_crank, n_grank, n_pp, n_acc, n_pc, n_count = statistics.country_rank, statistics.global_rank, \
                statistics.pp, statistics.hit_accuracy, statistics.play_count, statistics.total_hits
    else:
        n_crank, n_grank, n_pp, n_acc, n_pc, n_count = statistics.country_rank, statistics.global_rank, \
                                                       statistics.pp, statistics.hit_accuracy, \
                                                       statistics.play_count, statistics.total_hits
    # 新建
    im = Image.new('RGBA', (1000, 1350))
    draw = ImageDraw.Draw(im)
    # 获取背景
    bg_path = user_cache_path / str(info.id) / 'info.png'
    if bg_path.exists():
        bg = Image.open(bg_path)
    else:
        bg = await get_random_bg()
        if bg:
            bg = Image.open(BytesIO(bg))
    if bg:
        bg = bg.convert("RGBA")
        width, height = bg.size
        bg_ratio = height / width
        ratio = 1350 / 1000
        if bg_ratio > ratio:
            height = ratio * width
        else:
            width = height / ratio
        x, y = bg.size
        x, y = (x - width) // 2, (y - height) // 2
        bg = bg.crop((x, y, x + width, y + height)).resize((1000, 1350))
        im.alpha_composite(bg, (0, 0))
    # 获取头图，头像，地区，状态，supporter
    path = user_cache_path / str(info.id)
    if not path.exists():
        path.mkdir()
    country = osufile / 'flags' / f'{info.country_code}.png'
    # 底图
    im.alpha_composite(NewInfoImg)
    # 奖牌
    if info.badges:
        badges_num = len(info.badges)
        for num, badge in enumerate(info.badges):
            if badges_num <= 9:
                length = 50 + 100 * num
                height = 510
            elif num < 9:
                length = 50 + 100 * num
                height = 486
            else:
                length = 50 + 100 * (num - 9)
                height = 534
            badges_path = badge_cache_path / f'{hash(badge.description)}.png'
            if not badges_path.exists():
                await make_badge_cache_file(badge)
            badges_img = Image.open(badges_path).convert('RGBA').resize((86, 40))
            im.alpha_composite(badges_img, (length, height))
    # 地区
    country_bg = Image.open(country).convert('RGBA').resize((80, 54))
    im.alpha_composite(country_bg, (400, 394))
    # supporter
    if info.is_supporter:
        im.alpha_composite(SupporterBg.resize((54, 54)), (400, 280))
    # 经验
    if statistics.level.progress != 0:
        im.alpha_composite(ExpLeftBg, (50, 646))
        exp_width = statistics.level.progress * 7 - 3
        im.alpha_composite(ExpCenterBg.resize((exp_width, 10)), (54, 646))
        im.alpha_composite(ExpRightBg, (int(54 + exp_width), 646))
    # 模式
    draw.text((935, 50), GMN[mode], font=Torus_Regular_45, anchor='rm')
    # 玩家名
    draw.text((400, 205), info.username, font=Torus_Regular_50, anchor='lm')
    # 地区排名
    op, value = info_calc(statistics.country_rank, n_crank, rank=True)
    if not statistics.country_rank:
        t_crank = "#0"
    else:
        t_crank = f"#{statistics.country_rank:,}({op}{value:,})" \
            if value != 0 else f"#{statistics.country_rank:,}"
    draw.text((495, 448), t_crank, font=Torus_Regular_30, anchor='lb')
    # 等级
    draw.text((900, 650), str(statistics.level.current), font=Torus_Regular_25, anchor='mm')
    # 经验百分比
    draw.text((750, 660), f'{statistics.level.progress}%', font=Torus_Regular_20, anchor='rt')
    # 全球排名
    if not statistics.global_rank:
        draw.text((55, 785), "#0", font=Torus_Regular_35, anchor='lt')
    else:
        draw.text((55, 785), f"#{statistics.global_rank:,}", font=Torus_Regular_35, anchor='lt')
    op, value = info_calc(statistics.global_rank, n_grank, rank=True)
    if value != 0:
        draw.text((65, 820), f'{op}{value:,}', font=Torus_Regular_20, anchor='lt')
    # pp
    draw.text((295, 785), f'{statistics.pp:,}', font=Torus_Regular_35, anchor='lt')
    op, value = info_calc(statistics.pp, n_pp, pp=True)
    if value != 0:
        draw.text((305, 820), f'{op}{int(value)}', font=Torus_Regular_20)
    # SS - A
    # gc_x = 493
    for gc_num, (_, num) in enumerate(statistics.grade_counts):
        draw.text((493 + 100 * gc_num, 788), f'{num}', font=Torus_Regular_30, anchor='mt')
        # gc_x+=100
    # rank分
    draw.text((935, 895), f'{statistics.ranked_score:,}', font=Torus_Regular_40, anchor='rt')
    # acc
    op, value = info_calc(statistics.hit_accuracy, n_acc)
    t_acc = f'{statistics.hit_accuracy:.2f}%({op}{value:.2f}%)' if value != 0 else f'{statistics.hit_accuracy:.2f}%'
    draw.text((935, 965), t_acc, font=Torus_Regular_40, anchor='rt')
    # 游玩次数
    op, value = info_calc(statistics.play_count, n_pc)
    t_pc = f'{statistics.play_count:,}({op}{value:,})' if value != 0 else f'{statistics.play_count:,}'
    draw.text((935, 1035), t_pc, font=Torus_Regular_40, anchor='rt')
    # 总分
    draw.text((935, 1105), f'{statistics.total_score:,}', font=Torus_Regular_40, anchor='rt')
    # 总命中
    op, value = info_calc(statistics.total_hits, n_count)
    t_count = f'{statistics.total_hits:,}({op}{value:,})' if value != 0 else f'{statistics.total_hits:,}'
    draw.text((935, 1175), t_count, font=Torus_Regular_40, anchor='rt')
    # 游玩时间
    sec = timedelta(seconds=statistics.play_time)
    d_time = datetime(1, 1, 1) + sec
    t_time = "%dd %dh %dm %ds" % (sec.days, d_time.hour, d_time.minute, d_time.second)
    draw.text((935, 1245), t_time, font=Torus_Regular_40, anchor='rt')
    # 底部时间对比
    if day != 0 and user:
        day_delta = date.today() - user.date
        time = day_delta.days
        current_time = datetime.now().strftime('%Y/%m/%d %H:%M:%S')
        draw.text((260, 1305), current_time, font=Torus_Regular_25, anchor='la')
        text = f'| 数据对比于 {time} 天前'
        draw.text((515, 1305), text, font=Torus_Regular_25, anchor='la')
    else:
        current_time = datetime.now().strftime('%Y/%m/%d %H:%M:%S')
        draw.text((380, 1305), current_time, font=Torus_Regular_25, anchor='la')
    # 头像
    gif_frames = []
    user_icon = await open_user_icon(info)
    if not getattr(user_icon, "is_animated", False):
        icon_bg = user_icon.convert('RGBA').resize((300, 300))
        icon_img = draw_fillet(icon_bg, 25)
        im.alpha_composite(icon_img, (50, 148))
        byt = BytesIO()
        im.save(byt, "png")
        msg = MessageSegment.image(byt)
        im.close()
        user_icon.close()
        return msg
    for gif_frame in ImageSequence.Iterator(user_icon):
        # 将 GIF 图片中的每一帧转换为 RGBA 模式
        gif_frame = gif_frame.convert('RGBA').resize((300, 300))
        gif_frame = draw_fillet(gif_frame, 25)
        # 创建一个新的 RGBA 图片，将 PNG 图片作为背景，将当前帧添加到背景上
        rgba_frame = Image.new('RGBA', im.size, (0, 0, 0, 0))
        rgba_frame.paste(im, (0, 0), im)
        rgba_frame.paste(gif_frame, (50, 148), gif_frame)
        # 将 RGBA 图片转换为 RGB 模式，并添加到 GIF 图片中
        gif_frames.append(rgba_frame)
    gif_bytes = BytesIO()
    # 保存 GIF 图片
    gif_frames[0].save(gif_bytes, format='gif', save_all=True, append_images=gif_frames[1:])
    # 输出
    gif_frames[0].close()
    user_icon.close()
    msg = MessageSegment.image(gif_bytes)
    return msg



from functools import wraps
from typing import TypeVar, Callable
from typing_extensions import ParamSpec
from nonebot import logger


T = TypeVar('T')
P = ParamSpec('P')


def auto_retry(func: Callable[P, T]) -> Callable[P, T]:
    @wraps(func)
    async def wrapper(*args, **kwargs):
        for i in range(10):
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                logger.error(f"Retrying... {i + 1}/10")
                logger.exception(e)
        logger.error(f"多次重试失败，请检查网络连接")
    return wrapper



import asyncio
import random
from asyncio import TimerHandle
from difflib import SequenceMatcher
from io import BytesIO
from typing import Dict
from PIL import Image

from expiringdict import ExpiringDict
from nonebot import on_command, on_message
from nonebot.internal.rule import Rule
from nonebot.matcher import Matcher
from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent, MessageSegment
from nonebot.params import T_State
from .utils import split_msg
from ..info import get_bg
from ..utils import NGM
from ..api import osu_api
from ..schema import Score
from ..database.models import UserData

games: Dict[int, Score] = {}
pic_games: Dict[int, Score] = {}
timers: Dict[int, TimerHandle] = {}
pic_timers: Dict[int, TimerHandle] = {}
hint_dic = {'pic': False, 'artist': False, 'creator': False, 'mode': False}
pic_hint_dic = {'artist': False, 'creator': False, 'mode': False, 'audio': False}
group_hint = {}
pic_group_hint = {}
guess_audio = on_command('音频猜歌', priority=11, block=True)
guess_song_cache = ExpiringDict(1000, 60 * 60 * 24)


async def get_random_beatmap_set(binded_id, group_id, ttl=10):
    if ttl == 0:
        return
    selected_user = random.choice(binded_id)
    if not selected_user:
        return
    user = await UserData.filter(user_id=selected_user).first()
    bp_info = await osu_api('bp', user.osu_id, NGM[str(user.osu_mode)])
    if isinstance(bp_info, str):
        await guess_audio.finish('发生了错误，再试试吧')
    selected_score = random.choice([Score(**i) for i in bp_info])
    if selected_score.beatmapset.id not in guess_song_cache[group_id]:
        guess_song_cache[group_id].add(selected_score.beatmapset.id)
    else:
        return await get_random_beatmap_set(binded_id, group_id, ttl - 1)
    return selected_score, user


@guess_audio.handle(parameterless=[split_msg()])
async def _(state: T_State, event: GroupMessageEvent, bot: Bot, matcher: Matcher):
    if 'error' in state:
        await guess_audio.finish(MessageSegment.reply(event.message_id) + state['error'])
    mode = state['mode']
    group_id = event.group_id
    group_member = await bot.get_group_member_list(group_id=group_id)
    user_id_ls = [i['user_id'] for i in group_member]
    binded_id = await UserData.filter(user_id__in=user_id_ls).filter(osu_mode=mode).values_list('user_id', flat=True)
    if not binded_id:
        await guess_audio.finish('群里还没有人绑定该模式的osu账号呢，绑定了再来试试吧')
    if not guess_song_cache.get(group_id):
        guess_song_cache[group_id] = set()
    selected_score, selected_user = await get_random_beatmap_set(binded_id, group_id)
    if not selected_score:
        await guess_audio.finish('好像没有可以猜的歌了，今天的猜歌就到此结束吧！')
    if games.get(group_id, None):
        await guess_audio.finish('现在还有进行中的猜歌呢，请等待当前猜歌结束')
    games[group_id] = selected_score
    set_timeout(matcher, group_id)
    group_member = await bot.get_group_member_info(group_id=event.group_id, user_id=selected_user.user_id)
    name = group_member['card'] or group_member.get('nickname', '')
    await guess_audio.send(f'开始音频猜歌游戏，猜猜下面音频的曲名吧，该曲抽选自{name}的bp')
    print(selected_score.beatmapset.title)
    await guess_audio.finish(MessageSegment.record(f'https://cdn.sayobot.cn:25225/preview/{selected_score.beatmapset.id}.mp3'))


async def stop_game(matcher: Matcher, cid: int):
    timers.pop(cid, None)
    if games.get(cid, None):
        game = games.pop(cid)
        if group_hint.get(cid, None):
            group_hint[cid] = None
        msg = f"猜歌超时，游戏结束，正确答案是{game.beatmapset.title_unicode}"
        if game.beatmapset.title_unicode != game.beatmapset.title:
            msg += f' [{game.beatmapset.title}]'
        await matcher.send(msg)


async def pic_stop_game(matcher: Matcher, cid: int):
    pic_timers.pop(cid, None)
    if pic_games.get(cid, None):
        game = pic_games.pop(cid)
        if pic_group_hint.get(cid, None):
            pic_group_hint[cid] = None
        msg = f"猜歌超时，游戏结束，正确答案是{game.beatmapset.title_unicode}"
        if game.beatmapset.title_unicode != game.beatmapset.title:
            msg += f' [{game.beatmapset.title}]'
        await matcher.send(msg)

def set_timeout(matcher: Matcher, cid: int, timeout: float = 300):
    timer = timers.get(cid, None)
    if timer:
        timer.cancel()
    loop = asyncio.get_running_loop()
    timer = loop.call_later(
        timeout, lambda: asyncio.ensure_future(stop_game(matcher, cid))
    )
    timers[cid] = timer


def pic_set_timeout(matcher: Matcher, cid: int, timeout: float = 300):
    timer = pic_timers.get(cid, None)
    if timer:
        timer.cancel()
    loop = asyncio.get_running_loop()
    timer = loop.call_later(
        timeout, lambda: asyncio.ensure_future(pic_stop_game(matcher, cid))
    )
    pic_timers[cid] = timer


def game_running(event: GroupMessageEvent) -> bool:
    return bool(games.get(event.group_id, None))


def pic_game_running(event: GroupMessageEvent) -> bool:
    return bool(pic_games.get(event.group_id, None))


word_matcher = on_message(Rule(game_running), block=True, priority=12)


pic_word_matcher = on_message(Rule(pic_game_running), block=True, priority=12)


@word_matcher.handle()
async def _(event: GroupMessageEvent):
    song_name = games[event.group_id].beatmapset.title
    song_name_unicode = games[event.group_id].beatmapset.title_unicode
    r1 = SequenceMatcher(None, song_name.lower(), event.get_plaintext().lower()).ratio()
    r2 = SequenceMatcher(None, song_name_unicode.lower(), event.get_plaintext().lower()).ratio()
    if r1 >= 0.5 or r2 >= 0.5:
        games.pop(event.group_id)
        if group_hint.get(event.group_id, None):
            group_hint[event.group_id] = None
        msg = f"恭喜{event.sender.nickname}猜出正确答案为{song_name_unicode}"
        await word_matcher.finish(MessageSegment.reply(event.message_id) + msg)


@pic_word_matcher.handle()
async def _(event: GroupMessageEvent):
    song_name = pic_games[event.group_id].beatmapset.title
    song_name_unicode = pic_games[event.group_id].beatmapset.title_unicode
    r1 = SequenceMatcher(None, song_name.lower(), event.get_plaintext().lower()).ratio()
    r2 = SequenceMatcher(None, song_name_unicode.lower(), event.get_plaintext().lower()).ratio()
    if r1 >= 0.5 or r2 >= 0.5:
        pic_games.pop(event.group_id)
        if pic_group_hint.get(event.group_id, None):
            pic_group_hint[event.group_id] = None
        msg = f"恭喜{event.sender.nickname}猜出正确答案为{song_name_unicode}"
        await pic_word_matcher.finish(MessageSegment.reply(event.message_id) + msg)

hint = on_command('音频提示', priority=11, block=True, rule=Rule(game_running))


@hint.handle()
async def _(event: GroupMessageEvent):
    score = games[event.group_id]
    if not group_hint.get(event.group_id, None):
        group_hint[event.group_id] = hint_dic.copy()
    if all(group_hint[event.group_id].values()):
        await hint.finish('已无更多提示，加油哦')
    true_keys = []
    for key, value in group_hint[event.group_id].items():
        if not value:
            true_keys.append(key)
    action = random.choice(true_keys)
    if action == 'pic':
        group_hint[event.group_id]['pic'] = True
        await hint.finish(MessageSegment.image(score.beatmapset.covers.cover))
    if action == 'artist':
        group_hint[event.group_id]['artist'] = True
        msg = f'曲师为：{score.beatmapset.artist_unicode}'
        if score.beatmapset.artist_unicode != score.beatmapset.artist:
            msg += f' [{score.beatmapset.artist}]'
        await hint.finish(msg)
    if action == 'creator':
        group_hint[event.group_id]['creator'] = True
        await hint.finish(f'谱师为：{score.beatmapset.creator}')
    if action == 'mode':
        group_hint[event.group_id]['mode'] = True
        await hint.finish(f'模式为：{score.mode}')


pic_hint = on_command('图片提示', priority=11, block=True, rule=Rule(pic_game_running))


@pic_hint.handle()
async def _(event: GroupMessageEvent):
    score = pic_games[event.group_id]
    if not pic_group_hint.get(event.group_id, None):
        pic_group_hint[event.group_id] = pic_hint_dic.copy()
    if all(pic_group_hint[event.group_id].values()):
        await pic_hint.finish('已无更多提示，加油哦')
    true_keys = []
    for key, value in pic_group_hint[event.group_id].items():
        if not value:
            true_keys.append(key)
    action = random.choice(true_keys)
    if action == 'audio':
        pic_group_hint[event.group_id]['audio'] = True
        await pic_hint.finish(
            MessageSegment.record(f'https://cdn.sayobot.cn:25225/preview/{score.beatmapset.id}.mp3'))
    if action == 'artist':
        pic_group_hint[event.group_id]['artist'] = True
        msg = f'曲师为：{score.beatmapset.artist_unicode}'
        if score.beatmapset.artist_unicode != score.beatmapset.artist:
            msg += f' [{score.beatmapset.artist}]'
        await pic_hint.finish(msg)
    if action == 'creator':
        pic_group_hint[event.group_id]['creator'] = True
        await pic_hint.finish(f'谱师为：{score.beatmapset.creator}')
    if action == 'mode':
        pic_group_hint[event.group_id]['mode'] = True
        await pic_hint.finish(f'模式为：{score.mode}')


guess_pic = on_command('图片猜歌', priority=11, block=True)


@guess_pic.handle(parameterless=[split_msg()])
async def _(state: T_State, event: GroupMessageEvent, bot: Bot, matcher: Matcher):
    if 'error' in state:
        await guess_audio.finish(MessageSegment.reply(event.message_id) + state['error'])
    mode = state['mode']
    group_id = event.group_id
    group_member = await bot.get_group_member_list(group_id=group_id)
    user_id_ls = [i['user_id'] for i in group_member]
    binded_id = await UserData.filter(user_id__in=user_id_ls).filter(osu_mode=mode).values_list('user_id', flat=True)
    if not binded_id:
        await guess_pic.finish('群里还没有人绑定该模式的osu账号呢，绑定了再来试试吧')
    if not guess_song_cache.get(group_id):
        guess_song_cache[group_id] = set()
    selected_score, selected_user = await get_random_beatmap_set(binded_id, group_id)
    if not selected_score:
        await guess_pic.finish('好像没有可以猜的歌了，今天的猜歌就到此结束吧！')
    if pic_games.get(group_id, None):
        await guess_pic.finish('现在还有进行中的猜歌呢，请等待当前猜歌结束')
    pic_games[group_id] = selected_score
    pic_set_timeout(matcher, group_id)
    group_member = await bot.get_group_member_info(group_id=event.group_id, user_id=selected_user.user_id)
    name = group_member['card'] or group_member.get('nickname', '')
    await guess_pic.send(f'开始图片猜歌游戏，猜猜下面图片的曲名吧，该曲抽选自{name}的bp')
    byt = await get_bg(selected_score.beatmap.id)
    img = Image.open(byt)
    width, height = img.size
    crop_width = int(width * 0.3)
    crop_height = int(height * 0.3)
    left = random.randint(0, width - crop_width)
    top = random.randint(0, height - crop_height)
    right = left + crop_width
    bottom = top + crop_height
    cropped_image = img.crop((left, top, right, bottom))
    byt = BytesIO()
    cropped_image.save(byt, 'png')
    print(selected_score.beatmapset.title_unicode)
    await guess_pic.finish(MessageSegment.image(byt))



GM = {0: 'osu', 1: 'taiko', 2: 'fruits', 3: 'mania'}
NGM = {'0': 'osu', '1': 'taiko', '2': 'fruits', '3': 'mania'}
GMN = {'osu': 'Std', 'taiko': 'Taiko', 'fruits': 'Ctb', 'mania': 'Mania'}
FGM = {'osu': 0, 'taiko': 1, 'fruits': 2, 'mania': 3}


def mods2list(args: str) -> list:
    args = args.replace(' ', '').replace(',', '').replace('，', '')
    args = args.upper()
    return [args[i:i + 2] for i in range(0, len(args), 2)]
